"""
Similarity interface.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import numpy as np

import eta.core.utils as etau

import fiftyone.core.brain as fob
import fiftyone.core.utils as fou

fbs = fou.lazy_import("fiftyone.brain.internal.core.similarity")
fbu = fou.lazy_import("fiftyone.brain.internal.core.utils")


class SimilarityResults(fob.BrainResults):
    """Class storing the results of :meth:`fiftyone.brain.compute_similarity`.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`SimilarityConfig` used
        embeddings (None): a ``num_embeddings x num_dims`` array of embeddings
        sample_ids (None): a ``num_embeddings`` array of sample IDs
        label_ids (None): a ``num_embeddings`` array of label IDs, if
            applicable
    """

    def __init__(
        self, samples, config, embeddings=None, sample_ids=None, label_ids=None
    ):
        if embeddings is None:
            embeddings, sample_ids, label_ids = fbu.get_embeddings(
                samples._dataset,
                patches_field=config.patches_field,
                embeddings_field=config.embeddings_field,
            )
        elif sample_ids is None:
            sample_ids, label_ids = fbu.get_ids(
                samples,
                patches_field=config.patches_field,
                data=embeddings,
                data_type="embeddings",
            )

        self._samples = samples
        self._config = config
        self.embeddings = embeddings
        self.sample_ids = sample_ids
        self.label_ids = label_ids

        self._last_view = None
        self._curr_view = None
        self._curr_sample_ids = None
        self._curr_label_ids = None
        self._curr_keep_inds = None
        self._curr_good_inds = None
        self._neighbors_helper = None
        self._thresh = None
        self._unique_ids = None
        self._duplicate_ids = None
        self._neighbors_map = None
        self._model = None

        self.use_view(samples)

    def __enter__(self):
        self._last_view = self.view
        return self

    def __exit__(self, *args):
        self.use_view(self._last_view)
        self._last_view = None

    @property
    def config(self):
        """The :class:`SimilarityConfig` for the results."""
        return self._config

    @property
    def index_size(self):
        """The number of active data points in the index.

        If :meth:`use_view` has been called to restrict the index, this
        property will reflect the size of the active index.
        """
        return len(self._curr_sample_ids)

    @property
    def total_index_size(self):
        """The total number of data points in the index.

        If :meth:`use_view` has been called to restrict the index, this value
        may be larger than the current :meth:`index_size`.
        """
        return len(self.embeddings)

    @property
    def missing_size(self):
        """The total number of data points in :meth:`view` that are missing
        from this index.

        This property is only applicable when :meth:`use_view` has been called,
        and it will be ``None`` if no data points are missing.
        """
        good = self._curr_good_inds

        if good is None:
            return None

        return good.size - np.count_nonzero(good)

    @property
    def view(self):
        """The :class:`fiftyone.core.collections.SampleCollection` against
        which results are currently being generated.

        If :meth:`use_view` has been called, this view may be different than
        the collection on which the full index was generated.
        """
        return self._curr_view

    @property
    def thresh(self):
        """The threshold used by the last call to :meth:`find_duplicates` or
        :meth:`find_unique`.
        """
        return self._thresh

    @property
    def current_sample_ids(self):
        """The sample IDs of the currently active data points in the index.

        If :meth:`use_view` has been called, this may be a subset of the full
        index.
        """
        return self._curr_sample_ids

    @property
    def current_label_ids(self):
        """The label IDs of the currently active data points in the index, or
        ``None`` if not applicable.

        If :meth:`use_view` has been called, this may be a subset of the full
        index.
        """
        return self._curr_label_ids

    @property
    def unique_ids(self):
        """A list of unique IDs from the last call to :meth:`find_duplicates`
        or :meth:`find_unique`.
        """
        return self._unique_ids

    @property
    def duplicate_ids(self):
        """A list of duplicate IDs from the last call to
        :meth:`find_duplicates` or :meth:`find_unique`.
        """
        return self._duplicate_ids

    @property
    def neighbors_map(self):
        """A dictionary mapping IDs to lists of ``(dup_id, dist)`` tuples from
        the last call to :meth:`find_duplicates`.
        """
        return self._neighbors_map

    def use_view(
        self, sample_collection, allow_missing=True, warn_missing=False
    ):
        """Restricts the index to the provided view.

        Subsequent calls to methods on this instance will only contain results
        from the specified view rather than the full index.

        Use :meth:`clear_view` to reset to the full index. Or, equivalently,
        use the context manager interface as demonstrated below to
        automatically reset the view when the context exits.

        Example usage::

            import fiftyone as fo
            import fiftyone.brain as fob
            import fiftyone.zoo as foz

            dataset = foz.load_zoo_dataset("quickstart")

            results = fob.compute_similarity(dataset)
            print(results.index_size)  # 200

            view = dataset.take(50)

            with results.use_view(view):
                print(results.index_size)  # 50

                results.find_unique(10)
                print(results.unique_ids)

                plot = results.visualize_unique()
                plot.show()

        Args:
            sample_collection: a
                :class:`fiftyone.core.collections.SampleCollection`
            allow_missing (True): whether to allow the provided collection to
                contain data points that this index does not contain (True) or
                whether to raise an error in this case (False)
            warn_missing (False): whether to log a warning if the provided
                collection contains data points that this index does not
                contain

        Returns:
            self
        """
        sample_ids, label_ids, keep_inds, good_inds = fbu.filter_ids(
            sample_collection,
            self.sample_ids,
            self.label_ids,
            patches_field=self._config.patches_field,
            allow_missing=allow_missing,
            warn_missing=warn_missing,
        )

        self._curr_view = sample_collection
        self._curr_sample_ids = sample_ids
        self._curr_label_ids = label_ids
        self._curr_keep_inds = keep_inds
        self._curr_good_inds = good_inds

        return self

    def clear_view(self):
        """Clears the view set by :meth:`use_view`, if any.

        Subsequent operations will be performed on the full index.
        """
        self.use_view(self._samples)

    def values(self, path_or_expr):
        """Extracts a flat list of values from the given field or expression
        corresponding to the current :meth:`view`.

        This method always returns values in the same order as
        :meth:`current_sample_ids` and :meth:`current_label_ids`.

        Args:
            path_or_expr: the values to extract, which can be:

                -   the name of a sample field or ``embedded.field.name`` from
                    which to extract numeric or string values
                -   a :class:`fiftyone.core.expressions.ViewExpression`
                    defining numeric or string values to compute via
                    :meth:`fiftyone.core.collections.SampleCollection.values`

        Returns:
            a list of values
        """
        return fbs.values(self, path_or_expr)

    def plot_distances(self, bins=100, log=False, backend="plotly", **kwargs):
        """Plots a histogram of the distance between each example and its
        nearest neighbor.

        If `:meth:`find_duplicates` or :meth:`find_unique` has been executed,
        the threshold used is also indicated on the plot.

        Args:
            bins (100): the number of bins to use
            log (False): whether to use a log scale y-axis
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method

        Returns:
            one of the following:

            -   a :class:`fiftyone.core.plots.plotly.PlotlyNotebookPlot`, if
                you are working in a notebook context and the plotly backend is
                used
            -   a plotly or matplotlib figure, otherwise
        """
        return fbs.plot_distances(self, bins, log, backend, **kwargs)

    def sort_by_similarity(
        self,
        query,
        k=None,
        reverse=False,
        aggregation="mean",
        dist_field=None,
        _mongo=False,
    ):
        """Returns a view that sorts the samples/labels in :meth:`view` by
        similarity to the specified query.

        When querying by IDs, the query can be any ID(s) in the full index of
        this instance, even if the current :meth:`view` contains a subset of
        the full index.

        Args:
            query: the query, which can be any of the following:

                -   an ID or iterable of IDs
                -   a ``num_dims`` vector or ``num_queries x num_dims`` array
                    of vectors
                -   a prompt or iterable of prompts (if supported by the index)

            k (None): the number of matches to return. By default, all
                samples/labels are included
            reverse (False): whether to sort by least similarity
            aggregation ("mean"): the aggregation method to use to compute
                composite similarities. Only applicable when ``query`` contains
                multiple queries. Supported values are
                ``("mean", "min", "max")``
            dist_field (None): the name of a float field in which to store the
                distance of each example to the specified query. The field is
                created if necessary

        Returns:
            a :class:`fiftyone.core.view.DatasetView`
        """
        return fbs.sort_by_similarity(
            self, query, k, reverse, aggregation, dist_field, _mongo
        )

    def find_duplicates(self, thresh=None, fraction=None):
        """Queries the index to find near-duplicate examples based on the
        provided parameters.

        Calling this method populates the :meth:`unique_ids`,
        :meth:`duplicate_ids`, :attr:`neighbors_map`, and :attr:`thresh`
        properties of this object with the results of the query.

        Use :meth:`duplicates_view` and :meth:`visualize_duplicates` to analyze
        the results generated by this method.

        Args:
            thresh (None): a distance threshold to use to determine duplicates.
                If specified, the non-duplicate set will be the (approximately)
                largest set such that all pairwise distances between
                non-duplicate examples are greater than this threshold
            fraction (None): a desired fraction of images/patches to tag as
                duplicates, in ``[0, 1]``. In this case ``thresh`` is
                automatically tuned to achieve the desired fraction of
                duplicates
        """
        return fbs.find_duplicates(self, thresh, fraction)

    def find_unique(self, count):
        """Queries the index to select a subset of examples of the specified
        size that are maximally unique with respect to each other.

        Calling this method populates the :meth:`unique_ids`,
        :meth:`duplicate_ids`, and :attr:`thresh` properties of this object
        with the results of the query.

        Use :meth:`unique_view` and :meth:`visualize_unique` to analyze the
        results generated by this method.

        Args:
            count: the desired number of unique examples
        """
        return fbs.find_unique(self, count)

    def duplicates_view(
        self,
        type_field=None,
        id_field=None,
        dist_field=None,
        sort_by="distance",
        reverse=False,
    ):
        """Returns a view that contains only the duplicate examples and their
        corresponding nearest non-duplicate examples generated by the last call
        to :meth:`find_duplicates`.

        If you are analyzing patches, the returned view will be a
        :class:`fiftyone.core.patches.PatchesView`.

        The examples are organized so that each non-duplicate is immediately
        followed by all duplicate(s) that are nearest to it.

        Args:
            type_field (None): the name of a string field in which to store
                ``"nearest"`` and ``"duplicate"`` labels. The field is created
                if necessary
            id_field (None): the name of a string field in which to store the
                ID of the nearest non-duplicate for each example in the view.
                The field is created if necessary
            dist_field (None): the name of a float field in which to store the
                distance of each example to its nearest non-duplicate example.
                The field is created if necessary
            sort_by ("distance"): specifies how to sort the groups of duplicate
                examples. The supported values are:

                -   ``"distance"``: sort the groups by the distance between the
                    non-duplicate and its (nearest, if multiple) duplicate
                -   ``"count"``: sort the groups by the number of duplicate
                    examples

            reverse (False): whether to sort in descending order

        Returns:
            a :class:`fiftyone.core.view.DatasetView`
        """
        if self.neighbors_map is None:
            raise ValueError(
                "You must first call `find_duplicates()` to generate results"
            )

        return fbs.duplicates_view(
            self, type_field, id_field, dist_field, sort_by, reverse
        )

    def unique_view(self):
        """Returns a view that contains only the unique examples generated by
        the last call to :meth:`find_duplicates` or :meth:`find_unique`.

        If you are analyzing patches, the returned view will be a
        :class:`fiftyone.core.patches.PatchesView`.

        Returns:
            a :class:`fiftyone.core.view.DatasetView`
        """
        if self.unique_ids is None:
            raise ValueError(
                "You must first call `find_unique()` or `find_duplicates()` "
                "to generate results"
            )

        return fbs.unique_view(self)

    def visualize_duplicates(
        self, visualization=None, backend="plotly", **kwargs
    ):
        """Generates an interactive scatterplot of the results generated by the
        last call to :meth:`find_duplicates`.

        If provided, the ``visualization`` argument can be any visualization
        computed on the same dataset (or subset of it) as long as it contains
        every sample/object in the view whose results you are visualizing. If
        no ``visualization`` argument is provided and the embeddings
        have more than 3 dimensions, a 2D representation of the embeddings is
        computed via :meth:`fiftyone.brain.compute_visualization`.

        The points are colored based on the following partition:

            -   "duplicate": duplicate example
            -   "nearest": nearest neighbor of a duplicate example
            -   "unique": the remaining unique examples

        Edges are also drawn between each duplicate and its nearest
        non-duplicate neighbor.

        You can attach plots generated by this method to an App session via its
        :attr:`fiftyone.core.session.Session.plots` attribute, which will
        automatically sync the session's view with the currently selected
        points in the plot.

        Args:
            visualization (None): a
                :class:`fiftyone.brain.visualization.VisualizationResults`
                instance to use to visualize the results
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method:

                -   "plotly" backend: :meth:`fiftyone.core.plots.plotly.scatterplot`
                -   "matplotlib" backend: :meth:`fiftyone.core.plots.matplotlib.scatterplot`

        Returns:
            a :class:`fiftyone.core.plots.base.InteractivePlot`
        """
        if self.neighbors_map is None:
            raise ValueError(
                "You must first call `find_duplicates()` to generate results"
            )

        return fbs.visualize_duplicates(self, visualization, backend, **kwargs)

    def visualize_unique(self, visualization=None, backend="plotly", **kwargs):
        """Generates an interactive scatterplot of the results generated by the
        last call to :meth:`find_unique`.

        If provided, the ``visualization`` argument can be any visualization
        computed on the same dataset (or subset of it) as long as it contains
        every sample/object in the view whose results you are visualizing. If
        no ``visualization`` argument is provided and the embeddings
        have more than 3 dimensions, a 2D representation of the embeddings is
        computed via :meth:`fiftyone.brain.compute_visualization`.

        The points are colored based on the following partition:

            -   "unique": the unique examples
            -   "other": the other examples

        You can attach plots generated by this method to an App session via its
        :attr:`fiftyone.core.session.Session.plots` attribute, which will
        automatically sync the session's view with the currently selected
        points in the plot.

        Args:
            visualization (None): a
                :class:`fiftyone.brain.visualization.VisualizationResults`
                instance to use to visualize the results
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method:

                -   "plotly" backend: :meth:`fiftyone.core.plots.plotly.scatterplot`
                -   "matplotlib" backend: :meth:`fiftyone.core.plots.matplotlib.scatterplot`

        Returns:
            a :class:`fiftyone.core.plots.base.InteractivePlot`
        """
        if self.unique_ids is None:
            raise ValueError(
                "You must first call `find_unique()` to generate results"
            )

        return fbs.visualize_unique(self, visualization, backend, **kwargs)

    def attributes(self):
        attrs = super().attributes()

        if self.config.embeddings_field is not None:
            attrs = [
                attr
                for attr in attrs
                if attr not in ("embeddings", "sample_ids", "label_ids")
            ]

        return attrs

    @classmethod
    def _from_dict(cls, d, samples, config):
        embeddings = d.get("embeddings", None)
        if embeddings is not None:
            embeddings = np.array(embeddings)

        sample_ids = d.get("sample_ids", None)
        if sample_ids is not None:
            sample_ids = np.array(sample_ids)

        label_ids = d.get("label_ids", None)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return cls(
            samples,
            config,
            embeddings=embeddings,
            sample_ids=sample_ids,
            label_ids=label_ids,
        )


class SimilarityConfig(fob.BrainMethodConfig):
    """Similarity configuration.

    Args:
        embeddings_field (None): the sample field containing the embeddings,
            if one was provided
        model (None): the :class:`fiftyone.core.models.Model` or class name of
            the model that was used to compute embeddings, if one was provided
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        metric (None): the embedding distance metric used
        supports_prompts (False): whether this run supports prompt queries
    """

    def __init__(
        self,
        embeddings_field=None,
        model=None,
        patches_field=None,
        metric=None,
        supports_prompts=None,
        **kwargs,
    ):
        if model is not None and not etau.is_str(model):
            model = etau.get_class_name(model)

        self.embeddings_field = embeddings_field
        self.model = model
        self.patches_field = patches_field
        self.metric = metric
        self.supports_prompts = supports_prompts
        super().__init__(**kwargs)

    @property
    def method(self):
        return "similarity"

    @property
    def run_cls(self):
        run_cls_name = self.__class__.__name__[: -len("Config")]
        return getattr(fbs, run_cls_name)



"""
Uniqueness methods.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np
import sklearn.metrics as skm
import sklearn.neighbors as skn

import eta.core.utils as etau

import fiftyone.core.brain as fob
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.media as fom
import fiftyone.core.validation as fov

import fiftyone.brain.internal.core.utils as fbu
import fiftyone.brain.internal.models as fbm


logger = logging.getLogger(__name__)

_ALLOWED_ROI_FIELD_TYPES = (
    fol.Detection,
    fol.Detections,
    fol.Polyline,
    fol.Polylines,
)

_DEFAULT_MODEL = "simple-resnet-cifar10"
_DEFAULT_BATCH_SIZE = 16

_MAX_PRECOMPUTE_DISTS = 15000  # ~1.7GB to store distance matrix in-memory


def compute_uniqueness(
    samples,
    uniqueness_field,
    roi_field,
    embeddings,
    model,
    force_square,
    alpha,
    batch_size,
    num_workers,
    skip_failures,
):
    """See ``fiftyone/brain/__init__.py``."""

    #
    # Algorithm
    #
    # Uniqueness is computed based on a classification model.  Each sample is
    # embedded into a vector space based on the model. Then, we compute the
    # knn's (k is a parameter of the uniqueness function). The uniqueness is
    # then proportional to these distances. The intuition is that a sample is
    # unique when it is far from other samples in the set. This is different
    # than, say, "representativeness" which would stress samples that are core
    # to dense clusters of related samples.
    #

    fov.validate_image_collection(samples)

    if roi_field is not None:
        fov.validate_collection_label_fields(
            samples, roi_field, _ALLOWED_ROI_FIELD_TYPES
        )

    if model is None and embeddings is None:
        model = fbm.load_model(_DEFAULT_MODEL)
        if batch_size is None:
            batch_size = _DEFAULT_BATCH_SIZE

    if etau.is_str(embeddings):
        embeddings_field = fbu.parse_embeddings_field(
            samples,
            embeddings,
            patches_field=roi_field,
            allow_embedded=model is None,
        )
        embeddings = None
    else:
        embeddings_field = None

    config = UniquenessConfig(
        uniqueness_field,
        roi_field=roi_field,
        embeddings_field=embeddings_field,
        model=model,
    )
    brain_key = uniqueness_field
    brain_method = config.build()
    brain_method.ensure_requirements()
    brain_method.register_run(samples, brain_key)

    if roi_field is not None:
        # @todo experiment with mean(), max(), abs().max(), etc
        agg_fcn = lambda e: np.mean(e, axis=0)
    else:
        agg_fcn = None

    embeddings, sample_ids, _ = fbu.get_embeddings(
        samples,
        model=model,
        patches_field=roi_field,
        embeddings_field=embeddings_field,
        embeddings=embeddings,
        force_square=force_square,
        alpha=alpha,
        handle_missing="image",
        agg_fcn=agg_fcn,
        batch_size=batch_size,
        num_workers=num_workers,
        skip_failures=skip_failures,
    )

    logger.info("Computing uniqueness...")
    uniqueness = _compute_uniqueness(embeddings)

    # Ensure field exists, even if `uniqueness` is empty
    samples._dataset.add_sample_field(uniqueness_field, fof.FloatField)

    uniqueness = {_id: u for _id, u in zip(sample_ids, uniqueness)}
    if uniqueness:
        samples.set_values(uniqueness_field, uniqueness, key_field="id")

    brain_method.save_run_results(samples, brain_key, None)

    logger.info("Uniqueness computation complete")


def _compute_uniqueness(embeddings, metric="euclidean"):
    K = 3

    num_embeddings = len(embeddings)
    if num_embeddings <= K:
        return [1] * num_embeddings
    elif num_embeddings <= _MAX_PRECOMPUTE_DISTS:
        embeddings = skm.pairwise_distances(embeddings, metric=metric)
        metric = "precomputed"
    else:
        logger.info(
            "Computing neighbors for %d embeddings; this may take awhile...",
            num_embeddings,
        )

    # First column of dists and indices is self-distance
    knns = skn.NearestNeighbors(metric=metric).fit(embeddings)
    dists, _ = knns.kneighbors(embeddings, n_neighbors=K + 1)

    #
    # @todo experiment on which method for assessing uniqueness is best
    #
    # To get something going, for now, just take a weighted mean
    #
    weights = [0.6, 0.3, 0.1]
    sample_dists = np.mean(dists[:, 1:] * weights, axis=1)

    # Normalize to keep the user on common footing across datasets
    sample_dists /= sample_dists.max()

    return sample_dists


# @todo move to `fiftyone/brain/uniqueness.py`
# Don't do this hastily; `get_brain_info()` on existing datasets has this
# class's full path in it and may need migration
class UniquenessConfig(fob.BrainMethodConfig):
    def __init__(
        self,
        uniqueness_field,
        roi_field=None,
        embeddings_field=None,
        model=None,
        **kwargs,
    ):
        if model is not None and not etau.is_str(model):
            model = etau.get_class_name(model)

        self.uniqueness_field = uniqueness_field
        self.roi_field = roi_field
        self.embeddings_field = embeddings_field
        self.model = model
        super().__init__(**kwargs)

    @property
    def method(self):
        return "uniqueness"


class Uniqueness(fob.BrainMethod):
    def ensure_requirements(self):
        pass

    def get_fields(self, samples, brain_key):
        fields = [self.config.uniqueness_field]
        if self.config.roi_field is not None:
            fields.append(self.config.roi_field)

        if self.config.embeddings_field is not None:
            fields.append(self.config.embeddings_field)

        return fields

    def cleanup(self, samples, brain_key):
        uniqueness_field = self.config.uniqueness_field
        samples._dataset.delete_sample_fields(uniqueness_field, error_level=1)

    def _validate_run(self, samples, brain_key, existing_info):
        self._validate_fields_match(
            brain_key, "uniqueness_field", existing_info
        )



"""
Visualization methods.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np
import sklearn.decomposition as skd
import sklearn.manifold as skm

import eta.core.utils as etau

import fiftyone.core.brain as fob
import fiftyone.core.expressions as foe
import fiftyone.core.plots as fop
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov
import fiftyone.zoo as foz

from fiftyone.brain.visualization import (
    VisualizationResults,
    UMAPVisualizationConfig,
    TSNEVisualizationConfig,
    PCAVisualizationConfig,
    ManualVisualizationConfig,
)
import fiftyone.brain.internal.core.utils as fbu

umap = fou.lazy_import("umap")


logger = logging.getLogger(__name__)

_DEFAULT_MODEL = "mobilenet-v2-imagenet-torch"
_DEFAULT_BATCH_SIZE = None


def compute_visualization(
    samples,
    patches_field,
    embeddings,
    points,
    brain_key,
    num_dims,
    method,
    model,
    force_square,
    alpha,
    batch_size,
    num_workers,
    skip_failures,
    **kwargs,
):
    """See ``fiftyone/brain/__init__.py``."""

    fov.validate_collection(samples)

    if method == "manual" and points is None:
        raise ValueError(
            "You must provide your own `points` when `method='manual'`"
        )

    if points is not None:
        method = "manual"
        model = None
        embeddings = None
        embeddings_field = None
        num_dims = points.shape[1]
    elif model is None and embeddings is None:
        model = foz.load_zoo_model(_DEFAULT_MODEL)
        if batch_size is None:
            batch_size = _DEFAULT_BATCH_SIZE

    if etau.is_str(embeddings):
        embeddings_field = fbu.parse_embeddings_field(
            samples,
            embeddings,
            patches_field=patches_field,
            allow_embedded=model is None,
        )
        embeddings = None
    else:
        embeddings_field = None

    config = _parse_config(
        embeddings_field, model, patches_field, method, num_dims, **kwargs
    )

    brain_method = config.build()
    brain_method.ensure_requirements()

    if brain_key is not None:
        brain_method.register_run(samples, brain_key)

    if points is None:
        embeddings, sample_ids, label_ids = fbu.get_embeddings(
            samples,
            model=model,
            patches_field=patches_field,
            embeddings_field=embeddings_field,
            embeddings=embeddings,
            force_square=force_square,
            alpha=alpha,
            batch_size=batch_size,
            num_workers=num_workers,
            skip_failures=skip_failures,
        )

        logger.info("Generating visualization...")
        points = brain_method.fit(embeddings)
    else:
        sample_ids, label_ids = fbu.get_ids(
            samples,
            patches_field=patches_field,
            data=points,
            data_type="points",
        )

    results = VisualizationResults(
        samples,
        config,
        points,
        sample_ids=sample_ids,
        label_ids=label_ids,
    )

    brain_method.save_run_results(samples, brain_key, results)

    return results


def values(results, path_or_expr):
    samples = results.view
    patches_field = results.config.patches_field
    if patches_field is not None:
        ids = results.current_label_ids
    else:
        ids = results.current_sample_ids

    return fbu.get_values(
        samples, path_or_expr, ids, patches_field=patches_field
    )


def visualize(
    results,
    labels=None,
    sizes=None,
    classes=None,
    backend="plotly",
    **kwargs,
):
    points = results.current_points
    samples = results.view
    patches_field = results.config.patches_field
    good_inds = results._curr_good_inds
    if patches_field is not None:
        ids = results.current_label_ids
    else:
        ids = results.current_sample_ids

    if good_inds is not None:
        if labels is not None and not _is_expr(labels):
            labels = fbu.filter_values(
                labels, good_inds, patches_field=patches_field
            )

        if sizes is not None and not _is_expr(sizes):
            sizes = fbu.filter_values(
                sizes, good_inds, patches_field=patches_field
            )

    if labels is not None and _is_expr(labels):
        labels = fbu.get_values(
            samples, labels, ids, patches_field=patches_field
        )

    if sizes is not None and _is_expr(sizes):
        sizes = fbu.get_values(
            samples, sizes, ids, patches_field=patches_field
        )

    return fop.scatterplot(
        points,
        samples=samples,
        ids=ids,
        link_field=patches_field,
        labels=labels,
        sizes=sizes,
        classes=classes,
        backend=backend,
        **kwargs,
    )


def _is_expr(arg):
    return isinstance(arg, (foe.ViewExpression, dict))


class Visualization(fob.BrainMethod):
    def ensure_requirements(self):
        pass

    def fit(self, embeddings):
        raise NotImplementedError("subclass must implement fit()")

    def get_fields(self, samples, brain_key):
        fields = []
        if self.config.patches_field is not None:
            fields.append(self.config.patches_field)

        return fields

    def cleanup(self, samples, brain_key):
        pass


class UMAPVisualization(Visualization):
    def ensure_requirements(self):
        fou.ensure_package(
            "umap-learn>=0.5",
            error_msg=(
                "You must install the `umap-learn>=0.5` package in order to "
                "use UMAP-based visualization. This is recommended, as UMAP "
                "is awesome! If you do not wish to install UMAP, try "
                "`method='tsne'` instead"
            ),
        )

    def fit(self, embeddings):
        _umap = umap.UMAP(
            n_components=self.config.num_dims,
            n_neighbors=self.config.num_neighbors,
            metric=self.config.metric,
            min_dist=self.config.min_dist,
            random_state=self.config.seed,
            verbose=self.config.verbose,
        )
        return _umap.fit_transform(embeddings)


class TSNEVisualization(Visualization):
    def fit(self, embeddings):
        if self.config.pca_dims is not None:
            _pca = skd.PCA(
                n_components=self.config.pca_dims,
                svd_solver=self.config.svd_solver,
                random_state=self.config.seed,
            )
            embeddings = _pca.fit_transform(embeddings)

        embeddings = embeddings.astype(np.float32, copy=False)

        verbose = 2 if self.config.verbose else 0

        _tsne = skm.TSNE(
            n_components=self.config.num_dims,
            perplexity=self.config.perplexity,
            learning_rate=self.config.learning_rate,
            metric=self.config.metric,
            init="pca",  # "random" or "pca"
            n_iter=self.config.max_iters,
            random_state=self.config.seed,
            verbose=verbose,
        )
        return _tsne.fit_transform(embeddings)


class PCAVisualization(Visualization):
    def fit(self, embeddings):
        _pca = skd.PCA(
            n_components=self.config.num_dims,
            svd_solver=self.config.svd_solver,
            random_state=self.config.seed,
        )
        return _pca.fit_transform(embeddings)


class ManualVisualization(Visualization):
    def fit(self, embeddings):
        raise NotImplementedError(
            "The low-dimensional representation must be manually provided "
            "when using this method"
        )


def _parse_config(
    embeddings_field, model, patches_field, method, num_dims, **kwargs
):
    if method is None or method == "umap":
        config_cls = UMAPVisualizationConfig
    elif method == "tsne":
        config_cls = TSNEVisualizationConfig
    elif method == "pca":
        config_cls = PCAVisualizationConfig
    elif method == "manual":
        config_cls = ManualVisualizationConfig
    else:
        raise ValueError("Unsupported method '%s'" % method)

    return config_cls(
        embeddings_field=embeddings_field,
        model=model,
        patches_field=patches_field,
        num_dims=num_dims,
        **kwargs,
    )



"""
Simple configuration setup for these experiments

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import argparse


DATA_DIR = "/scratch/jason-data-cache/"
MODEL_DIR = "/scratch/jason-model-cache/"


class Config:
    """Encapsulates the configuration settings for a trial."""

    def __init__(self, d):
        """d is a dictionary of arguments from command-line."""

        self.batch_size = d["batch_size"]
        self.cold_start = d["cold_start"]
        self.epochs = d["epochs"]

        self.fixes = d["fixes"]
        self.make_fixes = True
        self.perfect_fixes = False
        if self.fixes == "no":
            self.makes_fixes = False
        if self.fixes == "perfect":
            self.make_fixes = True
            self.perfect_fixes = True

        self.model_path = d["model_path"]
        self.n_rounds = d["num_rounds"]
        self.n_max = d["num_max_samples"]
        self.p_corrupt = d["percent_corrupt"]
        self.p_fixable = d["percent_fixable"]
        self.p_initial = d["percent_initial"]

        self.runs = d["runs"]
        self.start_ipython = d["start_ipython"]
        self.stats_path = d["stats_path"]
        self.take = d["take"]

    def __str__(self):
        return str(vars(self))


def commandline():
    parser = argparse.ArgumentParser(add_help=True)

    parser.add_argument(
        "--batch_size",
        "-b",
        type=int,
        default=512,
        help="what is the batch size for the training iterations",
    )
    parser.add_argument(
        "--cold_start",
        "--cold-start",
        "-c",
        dest="cold_start",
        action="store_true",
    )
    parser.add_argument(
        "--no_cold_start",
        "--no-cold-start",
        "--no-cold_start",
        dest="cold_start",
        action="store_false",
    )
    parser.set_defaults(cold_start=False)
    parser.add_argument(
        "--epochs",
        "-e",
        type=int,
        default=24,
        help="what is the number of epochs for each training iteration",
    )
    parser.add_argument(
        "--fixes",
        "-f",
        type=str,
        default="yes",
        help="make fixes? options are yes, no, and perfect",
    )
    parser.add_argument(
        "--model_path",
        "-m",
        "--model-path",
        type=str,
        default=None,
        help="path to the file for loading/saving a model when applicable",
    )
    parser.add_argument(
        "--num_rounds",
        "--n_rounds",
        "--num-rounds",
        "--n-rounds",
        type=int,
        default=5,
        help="how many rounds to execute in a run",
    )
    parser.add_argument(
        "--num_max_samples",
        "--n_max_samples",
        "--num-max-samples",
        "--n-max-samples" "--num-max",
        "--num_max",
        "--n_max",
        "--n-max",
        type=int,
        default=-1,
        help="max samples to use in a run (default will use all samples)",
    )
    parser.add_argument(
        "--percent_corrupt",
        "--p_corrupt",
        "--percent-corrupt",
        "--p-corrupt",
        type=float,
        default=0.2,
        help="what percentage [0,1] of the data (annotations) to corrupt",
    )
    parser.add_argument(
        "--percent_fixable",
        "--p_fixable",
        "--percent-fixable",
        "--p-fixable",
        type=float,
        default=0.2,
        help="what percentage [0,1] of the data is fixable each round",
    )
    parser.add_argument(
        "--percent_initial",
        "--p_initial",
        "--percent-initial",
        "--p-initial",
        type=float,
        default=0.25,
        help="what percentage [0,1] of the data to use for initial training",
    )
    parser.add_argument(
        "--runs",
        "-r",
        type=int,
        default=6,
        help="how many times to run the trial for statistical purposes",
    )
    parser.add_argument(
        "--start_ipython",
        "--start-ipython",
        "-i",
        dest="start_ipython",
        action="store_true",
    )
    parser.add_argument(
        "--no_start_ipython",
        "--no-start-ipython",
        "--no-start_ipython",
        dest="start_ipython",
        action="store_false",
    )
    parser.set_defaults(start_ipython=False)
    parser.add_argument(
        "--stats_path",
        "-s",
        "--stats-path",
        type=str,
        default=None,
        help="path to the file for saving the statistics in json if desired",
    )
    parser.add_argument(
        "--take",
        "-t",
        type=int,
        default=None,
        help="how many samples to take in a subset of the dataset (default is all)",
    )

    args = parser.parse_args()

    return Config(args.__dict__)



"""
Definition and Implementation of a Simple Resnet and Supporting Functions

Original Implementation of this is from David Page's work on fast model
training with resnets.  <https://github.com/davidcpage/cifar10-fast/>

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import namedtuple, defaultdict
import copy
from functools import partial, singledispatch
from itertools import chain

import numpy as np
import torch
from torch import nn

from utils import *


torch.backends.cudnn.benchmark = True


## Data Preprocessing and Handling
def preprocess(dataset, transforms):
    dataset = copy.copy(dataset)  # shallow copy
    for transform in transforms:
        dataset["data"] = transform(dataset["data"])
    return dataset


@singledispatch
def normalise(x, mean, std):
    return (x - mean) / std


@normalise.register(np.ndarray)
def _(x, mean, std):
    # faster inplace for numpy arrays
    x = np.array(x, np.float32)
    x -= mean
    x *= 1.0 / std
    return x


unnormalise = lambda x, mean, std: x * std + mean


@singledispatch
def to_numpy(x):
    raise NotImplementedError


@to_numpy.register(torch.Tensor)
def _(x):
    return x.detach().cpu().numpy()


@singledispatch
def pad(x, border):
    raise NotImplementedError


@pad.register(np.ndarray)
def _(x, border):
    return np.pad(
        x, [(0, 0), (border, border), (border, border), (0, 0)], mode="reflect"
    )


@pad.register(torch.Tensor)
def _(x, border):
    return nn.ReflectionPad2d(border)(x)


@singledispatch
def transpose(x, source, target):
    raise NotImplementedError


@transpose.register(np.ndarray)
def _(x, source, target):
    return x.transpose([source.index(d) for d in target])


@transpose.register(torch.Tensor)
def _(x, source, target):
    return x.permute([source.index(d) for d in target])


class Crop(namedtuple("Crop", ("h", "w"))):
    def __call__(self, x, x0, y0):
        return x[..., y0 : y0 + self.h, x0 : x0 + self.w]

    def options(self, shape):
        *_, H, W = shape
        return [
            {"x0": x0, "y0": y0}
            for x0 in range(W + 1 - self.w)
            for y0 in range(H + 1 - self.h)
        ]

    def output_shape(self, shape):
        *_, H, W = shape
        return (*_, self.h, self.w)


@singledispatch
def flip_lr(x):
    raise NotImplementedError


@flip_lr.register(np.ndarray)
def _(x):
    return x[..., ::-1].copy()


@flip_lr.register(torch.Tensor)
def _(x):
    return torch.flip(x, [-1])


class FlipLR(namedtuple("FlipLR", ())):
    def __call__(self, x, choice):
        return flip_lr(x) if choice else x

    def options(self, shape):
        return [{"choice": b} for b in [True, False]]


class Cutout(namedtuple("Cutout", ("h", "w"))):
    def __call__(self, x, x0, y0):
        x[..., y0 : y0 + self.h, x0 : x0 + self.w] = 0.0
        return x

    def options(self, shape):
        *_, H, W = shape
        return [
            {"x0": x0, "y0": y0}
            for x0 in range(W + 1 - self.w)
            for y0 in range(H + 1 - self.h)
        ]


class Transform:
    def __init__(self, dataset, transforms):
        self.dataset, self.transforms = dataset, transforms
        self.choices = None

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        data, labels = self.dataset[index]
        data = data.copy()
        for choices, f in zip(self.choices, self.transforms):
            data = f(data, **choices[index])
        return data, labels

    def set_random_choices(self):
        self.choices = []
        x_shape = self.dataset[0][0].shape
        N = len(self)
        for t in self.transforms:
            self.choices.append(np.random.choice(t.options(x_shape), N))
            x_shape = (
                t.output_shape(x_shape)
                if hasattr(t, "output_shape")
                else x_shape
            )


## Utils


class PiecewiseLinear(namedtuple("PiecewiseLinear", ("knots", "vals"))):
    def __call__(self, t):
        return np.interp([t], self.knots, self.vals)[0]


class Const(namedtuple("Const", ["val"])):
    def __call__(self, x):
        return self.val


## Define the network

has_inputs = lambda node: type(node) is tuple


def build_graph(net):
    flattened = pipeline(net)
    resolve_input = (
        lambda rel_path, path, idx: normpath(sep.join((path, "..", rel_path)))
        if isinstance(rel_path, str)
        else flattened[idx + rel_path][0]
    )
    return {
        path: (
            node[0],
            [resolve_input(rel_path, path, idx) for rel_path in node[1]],
        )
        for idx, (path, node) in enumerate(flattened)
    }


def pipeline(net):
    return [
        (sep.join(path), (node if has_inputs(node) else (node, [-1])))
        for (path, node) in path_iter(net)
    ]


class Network(nn.Module):
    def __init__(self, net):
        super().__init__()
        self.graph = build_graph(net)
        for path, (val, _) in self.graph.items():
            setattr(self, path.replace("/", "_"), val)

    def nodes(self):
        return (node for node, _ in self.graph.values())

    def forward(self, inputs):
        outputs = dict(inputs)
        for k, (node, ins) in self.graph.items():
            # only compute nodes that are not supplied as inputs.
            if k not in outputs:
                outputs[k] = node(*[outputs[x] for x in ins])
        return outputs

    def half(self):
        for node in self.nodes():
            if isinstance(node, nn.Module) and not isinstance(
                node, nn.BatchNorm2d
            ):
                node.half()
        return self


class Identity(namedtuple("Identity", [])):
    def __call__(self, x):
        return x


class Add(namedtuple("Add", [])):
    def __call__(self, x, y):
        return x + y


class AddWeighted(namedtuple("AddWeighted", ["wx", "wy"])):
    def __call__(self, x, y):
        return self.wx * x + self.wy * y


class Mul(nn.Module):
    def __init__(self, weight):
        super().__init__()
        self.weight = weight

    def __call__(self, x):
        return x * self.weight


class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), x.size(1))


class Concat(nn.Module):
    def forward(self, *xs):
        return torch.cat(xs, 1)


class BatchNorm(nn.BatchNorm2d):
    def __init__(
        self,
        num_features,
        eps=1e-05,
        momentum=0.1,
        weight_freeze=False,
        bias_freeze=False,
        weight_init=1.0,
        bias_init=0.0,
    ):
        super().__init__(num_features, eps=eps, momentum=momentum)
        if weight_init is not None:
            self.weight.data.fill_(weight_init)
        if bias_init is not None:
            self.bias.data.fill_(bias_init)
        self.weight.requires_grad = not weight_freeze
        self.bias.requires_grad = not bias_freeze


def conv_bn(c_in, c_out):
    return {
        "conv": nn.Conv2d(
            c_in, c_out, kernel_size=3, stride=1, padding=1, bias=False
        ),
        "bn": BatchNorm(c_out),
        "relu": nn.ReLU(True),
    }


def residual(c):
    return {
        "in": Identity(),
        "res1": conv_bn(c, c),
        "res2": conv_bn(c, c),
        "add": (Add(), ["in", "res2/relu"]),
    }


def simple_resnet(
    channels=None,
    weight=0.125,
    pool=nn.MaxPool2d(2),
    extra_layers=(),
    res_layers=("layer1", "layer3"),
):
    channels = channels or {
        "prep": 64,
        "layer1": 128,
        "layer2": 256,
        "layer3": 512,
    }
    n = {
        "input": (None, []),
        "prep": conv_bn(3, channels["prep"]),
        "layer1": dict(
            conv_bn(channels["prep"], channels["layer1"]), pool=pool
        ),
        "layer2": dict(
            conv_bn(channels["layer1"], channels["layer2"]), pool=pool
        ),
        "layer3": dict(
            conv_bn(channels["layer2"], channels["layer3"]), pool=pool
        ),
        "pool": nn.MaxPool2d(4),
        "flatten": Flatten(),
        "linear": nn.Linear(channels["layer3"], 10, bias=False),
        "logits": Mul(weight),
    }
    for layer in res_layers:
        n[layer]["residual"] = residual(channels[layer])
    for layer in extra_layers:
        n[layer]["extra"] = conv_bn(channels[layer], channels[layer])
    return n


## Losses, Optimizers


class CrossEntropyLoss(namedtuple("CrossEntropyLoss", [])):
    def __call__(self, log_probs, target):
        return torch.nn.functional.nll_loss(
            log_probs, target, reduction="none"
        )


class KLLoss(namedtuple("KLLoss", [])):
    def __call__(self, log_probs):
        return -log_probs.mean(dim=1)


class Correct(namedtuple("Correct", [])):
    def __call__(self, classifier, target):
        return classifier.max(dim=1)[1] == target


class LogSoftmax(namedtuple("LogSoftmax", ["dim"])):
    def __call__(self, x):
        return torch.nn.functional.log_softmax(x, self.dim, _stacklevel=5)


x_ent_loss = Network(
    {
        "loss": (nn.CrossEntropyLoss(reduction="none"), ["logits", "target"]),
        "acc": (Correct(), ["logits", "target"]),
    }
)

label_smoothing_loss = lambda alpha: Network(
    {
        "logprobs": (LogSoftmax(dim=1), ["logits"]),
        "KL": (KLLoss(), ["logprobs"]),
        "xent": (CrossEntropyLoss(), ["logprobs", "target"]),
        "loss": (AddWeighted(wx=1 - alpha, wy=alpha), ["xent", "KL"]),
        "acc": (Correct(), ["logits", "target"]),
    }
)

trainable_params = lambda model: {
    k: p for k, p in model.named_parameters() if p.requires_grad
}


def nesterov_update(w, dw, v, lr, weight_decay, momentum):
    dw.add_(weight_decay, w).mul_(-lr)
    v.mul_(momentum).add_(dw)
    w.add_(dw.add_(momentum, v))


norm = lambda x: torch.norm(x.reshape(x.size(0), -1).float(), dim=1)[
    :, None, None, None
]


def LARS_update(w, dw, v, lr, weight_decay, momentum):
    nesterov_update(
        w,
        dw,
        v,
        lr * (norm(w) / (norm(dw) + 1e-2)).to(w.dtype),
        weight_decay,
        momentum,
    )


def zeros_like(weights):
    return [torch.zeros_like(w) for w in weights]


def optimiser(weights, param_schedule, update, state_init):
    weights = list(weights)
    return {
        "update": update,
        "param_schedule": param_schedule,
        "step_number": 0,
        "weights": weights,
        "opt_state": state_init(weights),
    }


def opt_step(update, param_schedule, step_number, weights, opt_state):
    step_number += 1
    param_values = {k: f(step_number) for k, f in param_schedule.items()}
    for w, v in zip(weights, opt_state):
        if w.requires_grad:
            update(w.data, w.grad.data, v, **param_values)
    return {
        "update": update,
        "param_schedule": param_schedule,
        "step_number": step_number,
        "weights": weights,
        "opt_state": opt_state,
    }


LARS = partial(optimiser, update=LARS_update, state_init=zeros_like)
SGD = partial(optimiser, update=nesterov_update, state_init=zeros_like)


## Training Code

MODEL = "model"
LOSS = "loss"
VALID_MODEL = "valid_model"
OPTS = "optimisers"
OUTPUT = "output"
ACT_LOG = "activation_log"


def reduce(batches, state, steps):
    # state: is a dictionary
    # steps: are functions that take (batch, state)
    # and return a dictionary of updates to the state (or None)

    for batch in chain(batches, [None]):
        # we send an extra batch=None at the end for steps that
        # need to do some tidying-up (e.g. log_activations)
        for step in steps:
            updates = step(batch, state)
            if updates:
                for k, v in updates.items():
                    state[k] = v
    return state


def forward(training_mode):
    def step(batch, state):
        if not batch:
            return
        model = (
            state[MODEL]
            if training_mode or (VALID_MODEL not in state)
            else state[VALID_MODEL]
        )
        if model.training != training_mode:  # without the guard it's slow!
            model.train(training_mode)
        return {OUTPUT: state[LOSS](model(batch))}

    return step


def backward(dtype=None):
    def step(batch, state):
        state[MODEL].zero_grad()
        if not batch:
            return
        loss = state[OUTPUT][LOSS]
        if dtype is not None:
            loss = loss.to(dtype)
        loss.sum().backward()

    return step


def opt_steps(batch, state):
    if not batch:
        return
    return {OPTS: [opt_step(**opt) for opt in state[OPTS]]}


def log_activations(node_names=("loss", "acc")):
    def step(batch, state):
        if "_tmp_logs_" not in state:
            state["_tmp_logs_"] = []
        if batch:
            state["_tmp_logs_"].extend(
                (k, state[OUTPUT][k].detach()) for k in node_names
            )
        else:
            res = {
                k: to_numpy(torch.cat(xs)).astype(np.float)
                for k, xs in group_by_key(state["_tmp_logs_"]).items()
            }
            del state["_tmp_logs_"]
            return {ACT_LOG: res}

    return step


epoch_stats = lambda state: {k: np.mean(v) for k, v in state[ACT_LOG].items()}

default_train_steps = (
    forward(training_mode=True),
    log_activations(("loss", "acc")),
    backward(),
    opt_steps,
)
default_valid_steps = (
    forward(training_mode=False),
    log_activations(("loss", "acc")),
)


def train_epoch(
    state,
    timer,
    train_batches,
    valid_batches,
    train_steps=default_train_steps,
    valid_steps=default_valid_steps,
    on_epoch_end=(lambda state: state),
):
    train_summary, train_time = (
        epoch_stats(on_epoch_end(reduce(train_batches, state, train_steps))),
        timer(),
    )
    valid_summary, valid_time = (
        epoch_stats(reduce(valid_batches, state, valid_steps)),
        timer(include_in_total=False),
    )  # DAWNBench rules
    return {
        "train": union({"time": train_time}, train_summary),
        "valid": union({"time": valid_time}, valid_summary),
        "total time": timer.total_time,
    }



"""
le_test2 --> initial tests for using fiftyone with this code
Loads the data, assigns ground-truth labels, then trains a clean model

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from functools import partial
import json
import os
import time

from imageio import imsave

import fiftyone as fo

from config import *
from datasets import *
from simple_resnet import *
from utils import Timer


TEMP_TRAIN_DIR = "/tmp/le_test/train"
TEMP_VALID_DIR = "/tmp/le_test/valid"

localtime = lambda: time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


def write_images(root, images, overwrite=False):
    paths = []
    for i, s in enumerate(images):
        path = os.path.join(root, f"{i:05d}.jpg")
        paths.append(path)

        if overwrite or not os.path.exists(path):
            img = s.copy()
            img = transpose(img, source="CHW", target="HWC")
            imsave(path, img)

    return paths


def main(config):
    #
    # Load dataset
    #
    # `train_set` and `valid_set` that are lists of `(image, label)` tuples
    #

    timer = Timer()
    whole_dataset = cifar10(root=DATA_DIR)
    print("Preprocessing training data")
    transforms = [
        partial(
            normalise,
            mean=np.array(cifar10_mean, dtype=np.float32),
            std=np.array(cifar10_std, dtype=np.float32),
        ),
        partial(transpose, source="NHWC", target="NCHW"),
    ]
    whole_train_set = list(
        zip(
            *preprocess(
                whole_dataset["train"], [partial(pad, border=4)] + transforms
            ).values()
        )
    )
    valid_set = list(
        zip(*preprocess(whole_dataset["valid"], transforms).values())
    )
    print(f"Finished loading and preprocessing in {timer():.2f} seconds")

    print(f"train set: {len(whole_train_set)} samples")
    print(f"valid set: {len(valid_set)} samples")

    # Write raw dataset to disk
    os.makedirs(TEMP_TRAIN_DIR, exist_ok=True)
    os.makedirs(TEMP_VALID_DIR, exist_ok=True)
    train_image_paths = write_images(
        TEMP_TRAIN_DIR, list(zip(*whole_train_set))[0]
    )
    valid_image_paths = write_images(TEMP_VALID_DIR, list(zip(*valid_set))[0])

    #
    # Load data into FiftyOne
    #

    timer = Timer()
    dataset = fo.Dataset("le_cifar10")

    # Train split
    train_samples = []
    for i, s in enumerate(whole_train_set):
        train_samples.append(
            fo.Sample(
                train_image_paths[i],
                tags=["train"],
                ground_truth=fo.Classification(label=cifar10_classes[s[1]]),
            )
        )

    train_ids = dataset.add_samples(train_samples)

    # Valid split
    valid_samples = []
    for i, s in enumerate(valid_set):
        valid_samples.append(
            fo.Sample(
                valid_image_paths[i],
                tags=["valid"],
                ground_truth=fo.Classification(label=cifar10_classes[s[1]]),
            )
        )

    valid_ids = dataset.add_samples(valid_samples)

    print(f"Finished getting data into fiftyone in {timer():.2f} seconds")
    print(dataset.summary())

    # function of dataset
    N_labels = 10

    # set up the variables for training the model in each increment of the
    # dataset size
    lr_schedule = PiecewiseLinear([0, 5, config.epochs], [0, 0.4, 0])
    train_transforms = [Crop(32, 32), FlipLR(), Cutout(8, 8)]

    # compute the derived parameters for the trial based on the dataset and the
    # provided config.
    total_N = len(whole_train_set)

    # should be cleaner
    if config.n_max < 0:
        config.n_max = total_N

    start_N = round(config.p_initial * total_N)

    incr_N = (
        0
        if config.n_rounds == 1
        else round((config.n_max - start_N) / (config.n_rounds - 1))
    )

    corrupt_N = round(config.p_corrupt * total_N)

    print(f"Setting up the experiment: {total_N} training samples.")
    print(f"- starting with {start_N}")
    print(f"- incrementing by {incr_N} for each of {config.n_rounds-1} rounds")
    print(f"- total rounds: {config.n_rounds}")

    print(f"Starting the model training at {localtime()}")

    inuse_N = start_N

    model = Network(simple_resnet()).to(device).half()
    logs, state = Table(), {MODEL: model, LOSS: x_ent_loss}

    valid_batches = DataLoader(
        valid_set, config.batch_size, shuffle=False, drop_last=False
    )

    # initially randomly shuffle the dataset and take the initial number of
    # samples
    whole_train_set_use = whole_train_set[0:inuse_N]
    whole_train_set_avail = whole_train_set[inuse_N:]
    print(
        f"Split training set into two; using {len(whole_train_set_use)}, "
        + f"available {len(whole_train_set_avail)}"
    )

    sm = torch.nn.Softmax(dim=1)

    stats = {}

    for iteration in range(config.n_rounds):
        print(f"beginning next round of training, using {inuse_N} samples")

        if config.cold_start:
            model = Network(simple_resnet()).to(device).half()
            logs, state = Table(), {MODEL: model, LOSS: x_ent_loss}

        train_batches = DataLoader(
            Transform(whole_train_set_use, train_transforms),
            config.batch_size,
            shuffle=True,
            set_random_choices=True,
            drop_last=True,
        )
        lr = (
            lambda step: lr_schedule(step / len(train_batches))
            / config.batch_size
        )
        opts = [
            SGD(
                trainable_params(model).values(),
                {
                    "lr": lr,
                    "weight_decay": Const(5e-4 * config.batch_size),
                    "momentum": Const(0.9),
                },
            )
        ]
        state[OPTS] = opts

        for epoch in range(config.epochs):
            logs.append(
                union(
                    {"epoch": epoch + 1},
                    train_epoch(
                        state,
                        Timer(torch.cuda.synchronize),
                        train_batches,
                        valid_batches,
                    ),
                )
            )
        logs.df().query(f"epoch=={config.epochs}")[
            ["train_acc", "valid_acc"]
        ].describe()

        model.train(False)  # == model.eval()

        # record scores for this iteration
        iteration_stats = {}
        iteration_stats["in_use"] = inuse_N

        correct = 0
        total = 0
        class_correct = list(0.0 for i in range(10))
        class_total = list(0.0 for i in range(10))
        with torch.no_grad():
            for data in valid_batches.dataloader:
                images, labels = data
                inputs = dict(input=images.cuda().half())
                outputs = model(inputs)
                y = outputs["logits"]
                _, predicted = torch.max(y, 1)
                total += labels.size(0)
                labels_gpu = labels.cuda().half()
                correct += (predicted == labels_gpu).sum().item()
                c = (predicted == labels_gpu).squeeze()
                for i in range(min(config.batch_size, len(labels))):
                    label = labels[i]
                    class_correct[label] += c[i].item()
                    class_total[label] += 1

        iteration_stats["validation_accuracy"] = correct / total

        model.train(True)

        # extend the corr_train_set_use with that from avail
        whole_train_set_use.extend(whole_train_set_avail[0:incr_N])
        whole_train_set_avail = whole_train_set_avail[incr_N:]
        inuse_N += incr_N
        assert inuse_N == len(whole_train_set_use)

        stats[inuse_N] = iteration_stats

    print(f"finished the full training; stats to follow")
    print(stats)

    if config.stats_path:
        with open(config.stats_path, "w") as fp:
            json.dump(stats, fp)

    if config.model_path:
        torch.save(model.state_dict(), config.model_path)


if __name__ == "__main__":

    config = commandline()

    print(f"running with config: {config}")

    main(config)



"""
Implementation of datasets for the experiments

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from functools import lru_cache as cache

import torch

from utils import *


class DataLoader:
    def __init__(
        self,
        dataset,
        batch_size,
        shuffle,
        set_random_choices=False,
        num_workers=0,
        drop_last=False,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.set_random_choices = set_random_choices
        self.dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=True,
            shuffle=shuffle,
            drop_last=drop_last,
        )

    def __iter__(self):
        if self.set_random_choices:
            self.dataset.set_random_choices()
        return (
            {"input": x.to(device).half(), "target": y.to(device).long()}
            for (x, y) in self.dataloader
        )

    def __len__(self):
        return len(self.dataloader)


@cache(None)
def cifar10(root="./data"):
    try:
        import torchvision

        download = lambda train: torchvision.datasets.CIFAR10(
            root=root, train=train, download=True
        )
        return {
            k: {"data": v.data, "targets": v.targets}
            for k, v in [
                ("train", download(train=True)),
                ("valid", download(train=False)),
            ]
        }
    except ImportError:
        from tensorflow.keras import datasets

        (
            (train_images, train_labels),
            (valid_images, valid_labels),
        ) = datasets.cifar10.load_data()
        return {
            "train": {"data": train_images, "targets": train_labels.squeeze()},
            "valid": {"data": valid_images, "targets": valid_labels.squeeze()},
        }


cifar10_mean, cifar10_std = [
    (
        125.31,
        122.95,
        113.87,
    ),  # equals np.mean(cifar10()['train']['data'], axis=(0,1,2))
    (
        62.99,
        62.09,
        66.70,
    ),  # equals np.std(cifar10()['train']['data'], axis=(0,1,2))
]
cifar10_classes = "airplane, automobile, bird, cat, deer, dog, frog, horse, ship, truck".split(
    ", "
)



"""
Uniqueness tests.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os
import unittest

import eta.core.storage as etas
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.brain as fob
import fiftyone.zoo as foz


def test_uniqueness():
    dataset = foz.load_zoo_dataset("cifar10", split="test")
    assert "uniqueness" not in dataset.get_field_schema()

    view = dataset.view().take(100)
    fob.compute_uniqueness(view)

    print(dataset)
    assert "uniqueness" in dataset.get_field_schema()


def test_gray():
    """Test default support for handling grayscale images.

    Requires Voxel51 Google Drive credentials to download the test data.
    """
    with etau.TempDir() as tmpdir:
        tmp_zip = os.path.join(tmpdir, "data.zip")
        tmp_data = os.path.join(tmpdir, "brain_grayscale_test_data")
        client = etas.GoogleDriveStorageClient()
        client.download("1ECeNnLmKQCHxlVdRqGefV5eXOD_OkmWx", tmp_zip)
        etau.extract_zip(tmp_zip, delete_zip=True)

        dataset = fo.Dataset.from_dir(tmp_data, fo.types.ImageDirectory)

        fob.compute_uniqueness(dataset)
        print(dataset)


if __name__ == "__main__":
    fo.config.show_progress_bars = True
    unittest.main(verbosity=2)



"""
Training functions

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import namedtuple
from functools import partial, singledispatch
from itertools import chain

import numpy as np
import torch

from fiftyone.brain.internal.models.simple_resnet import *

from utils import *


@singledispatch
def to_numpy(x):
    raise NotImplementedError


@to_numpy.register(torch.Tensor)
def _(x):
    return x.detach().cpu().numpy()


## Losses, Optimizers


class CrossEntropyLoss(namedtuple("CrossEntropyLoss", [])):
    def __call__(self, log_probs, target):
        return torch.nn.functional.nll_loss(
            log_probs, target, reduction="none"
        )


class KLLoss(namedtuple("KLLoss", [])):
    def __call__(self, log_probs):
        return -log_probs.mean(dim=1)


class Correct(namedtuple("Correct", [])):
    def __call__(self, classifier, target):
        return classifier.max(dim=1)[1] == target


class LogSoftmax(namedtuple("LogSoftmax", ["dim"])):
    def __call__(self, x):
        return torch.nn.functional.log_softmax(x, self.dim, _stacklevel=5)


x_ent_loss = Network(
    {
        "loss": (nn.CrossEntropyLoss(reduction="none"), ["logits", "target"]),
        "acc": (Correct(), ["logits", "target"]),
    }
)

label_smoothing_loss = lambda alpha: Network(
    {
        "logprobs": (LogSoftmax(dim=1), ["logits"]),
        "KL": (KLLoss(), ["logprobs"]),
        "xent": (CrossEntropyLoss(), ["logprobs", "target"]),
        "loss": (AddWeighted(wx=1 - alpha, wy=alpha), ["xent", "KL"]),
        "acc": (Correct(), ["logits", "target"]),
    }
)

trainable_params = lambda model: (
    {k: p for k, p in model.named_parameters() if p.requires_grad}
)


def nesterov_update(w, dw, v, lr, weight_decay, momentum):
    dw.add_(weight_decay, w).mul_(-lr)
    v.mul_(momentum).add_(dw)
    w.add_(dw.add_(momentum, v))


norm = lambda x: torch.norm(x.reshape(x.size(0), -1).float(), dim=1)[
    :, None, None, None
]


def LARS_update(w, dw, v, lr, weight_decay, momentum):
    nesterov_update(
        w,
        dw,
        v,
        lr * (norm(w) / (norm(dw) + 1e-2)).to(w.dtype),
        weight_decay,
        momentum,
    )


def zeros_like(weights):
    return [torch.zeros_like(w) for w in weights]


def optimiser(weights, param_schedule, update, state_init):
    weights = list(weights)
    return {
        "update": update,
        "param_schedule": param_schedule,
        "step_number": 0,
        "weights": weights,
        "opt_state": state_init(weights),
    }


def opt_step(update, param_schedule, step_number, weights, opt_state):
    step_number += 1
    param_values = {k: f(step_number) for k, f in param_schedule.items()}
    for w, v in zip(weights, opt_state):
        if w.requires_grad:
            update(w.data, w.grad.data, v, **param_values)

    return {
        "update": update,
        "param_schedule": param_schedule,
        "step_number": step_number,
        "weights": weights,
        "opt_state": opt_state,
    }


LARS = partial(optimiser, update=LARS_update, state_init=zeros_like)
SGD = partial(optimiser, update=nesterov_update, state_init=zeros_like)


## Training Code
LOSS = "loss"
OPTS = "optimisers"
ACT_LOG = "activation_log"


def reduce(batches, state, steps):
    # state: is a dictionary
    # steps: are functions that take (batch, state)
    # and return a dictionary of updates to the state (or None)

    for batch in chain(batches, [None]):
        # we send an extra batch=None at the end for steps that
        # need to do some tidying-up (e.g. log_activations)
        for step in steps:
            updates = step(batch, state)
            if updates:
                for k, v in updates.items():
                    state[k] = v

    return state


def forward(training_mode):
    def step(batch, state):
        if not batch:
            return

        model = (
            state[MODEL]
            if training_mode or (VALID_MODEL not in state)
            else state[VALID_MODEL]
        )
        if model.training != training_mode:  # without the guard it's slow!
            model.train(training_mode)

        return {OUTPUT: state[LOSS](model(batch))}

    return step


def backward(dtype=None):
    def step(batch, state):
        state[MODEL].zero_grad()
        if not batch:
            return

        loss = state[OUTPUT][LOSS]
        if dtype is not None:
            loss = loss.to(dtype)

        loss.sum().backward()

    return step


def opt_steps(batch, state):
    if not batch:
        return

    return {OPTS: [opt_step(**opt) for opt in state[OPTS]]}


def log_activations(node_names=("loss", "acc")):
    def step(batch, state):
        if "_tmp_logs_" not in state:
            state["_tmp_logs_"] = []
        if batch:
            state["_tmp_logs_"].extend(
                (k, state[OUTPUT][k].detach()) for k in node_names
            )
        else:
            res = {
                k: to_numpy(torch.cat(xs)).astype(np.float)
                for k, xs in group_by_key(state["_tmp_logs_"]).items()
            }
            del state["_tmp_logs_"]
            return {ACT_LOG: res}

    return step


epoch_stats = lambda state: {k: np.mean(v) for k, v in state[ACT_LOG].items()}

default_train_steps = (
    forward(training_mode=True),
    log_activations(("loss", "acc")),
    backward(),
    opt_steps,
)
default_valid_steps = (
    forward(training_mode=False),
    log_activations(("loss", "acc")),
)


def train_epoch(
    state,
    timer,
    train_batches,
    valid_batches,
    train_steps=default_train_steps,
    valid_steps=default_valid_steps,
    on_epoch_end=(lambda state: state),
):
    train_summary, train_time = (
        epoch_stats(on_epoch_end(reduce(train_batches, state, train_steps))),
        timer(),
    )
    valid_summary, valid_time = (
        epoch_stats(reduce(valid_batches, state, valid_steps)),
        timer(include_in_total=False),
    )  # DAWNBench rules
    return {
        "train": union({"time": train_time}, train_summary),
        "valid": union({"time": valid_time}, valid_summary),
        "total time": timer.total_time,
    }



"""
Duplicates methods.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
import itertools
import logging
import multiprocessing

import fiftyone.core.media as fom
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov


logger = logging.getLogger(__name__)


def compute_exact_duplicates(samples, num_workers, skip_failures):
    """See ``fiftyone/brain/__init__.py``."""

    fov.validate_collection(samples)

    if num_workers is None:
        if samples.media_type == fom.VIDEO:
            num_workers = multiprocessing.cpu_count()
        else:
            num_workers = 1

    logger.info("Computing filehashes...")

    method = "md5" if samples.media_type == fom.VIDEO else None

    if num_workers <= 1:
        hashes = _compute_filehashes(samples, method)
    else:
        hashes = _compute_filehashes_multi(samples, method, num_workers)

    num_missing = sum(h is None for h in hashes)
    if num_missing > 0:
        msg = "Failed to compute %d filehashes" % num_missing
        if skip_failures:
            logger.warning(msg)
        else:
            raise ValueError(msg)

    neighbors_map = defaultdict(list)

    observed_hashes = {}
    for _id, _hash in hashes.items():
        if _hash is None:
            continue

        if _hash in observed_hashes:
            neighbors_map[observed_hashes[_hash]].append(_id)
        else:
            observed_hashes[_hash] = _id

    return dict(neighbors_map)


def _compute_filehashes(samples, method):
    ids, filepaths = samples.values(["id", "filepath"])

    with fou.ProgressBar(total=len(ids)) as pb:
        return {
            _id: _compute_filehash(filepath, method)
            for _id, filepath in pb(zip(ids, filepaths))
        }


def _compute_filehashes_multi(samples, method, num_workers):
    ids, filepaths = samples.values(["id", "filepath"])

    methods = itertools.repeat(method)

    inputs = list(zip(ids, filepaths, methods))

    with fou.ProgressBar(total=len(inputs)) as pb:
        with multiprocessing.Pool(processes=num_workers) as pool:
            return {
                k: v
                for k, v in pb(
                    pool.imap_unordered(_do_compute_filehash, inputs)
                )
            }


def _compute_filehash(filepath, method):
    try:
        filehash = fou.compute_filehash(filepath, method=method)
    except:
        filehash = None

    return filehash


def _do_compute_filehash(args):
    _id, filepath, method = args
    try:
        filehash = fou.compute_filehash(filepath, method=method)
    except:
        filehash = None

    return _id, filehash



"""
Similarity tests.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import unittest

import fiftyone as fo
import fiftyone.brain as fob
import fiftyone.zoo as foz
from fiftyone import ViewField as F


def test_images():
    dataset = _load_images_dataset()

    results = dataset.load_brain_results("img_sim")

    assert results.total_index_size == len(dataset)
    assert set(dataset.values("id")) == set(results.sample_ids)


def test_images_subset():
    dataset = _load_images_dataset()

    results = dataset.load_brain_results("img_sim")

    view = dataset.take(10)
    results.use_view(view)

    assert results.index_size == len(view)
    assert set(view.values("id")) == set(results.current_sample_ids)


def test_images_missing():
    dataset = _load_images_dataset().limit(4).clone()
    dataset.add_samples(
        [
            fo.Sample(filepath="non-existent1.png"),
            fo.Sample(filepath="non-existent2.png"),
            fo.Sample(filepath="non-existent3.png"),
            fo.Sample(filepath="non-existent4.png"),
        ]
    )

    sample_ids = dataset[:4].values("id")

    results = fob.compute_similarity(dataset, batch_size=1)

    assert results.total_index_size == 4
    assert set(sample_ids) == set(results.sample_ids)

    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    results = fob.compute_similarity(
        dataset,
        model=model,
        embeddings="embeddings_missing",
        batch_size=1,
    )

    assert len(dataset.exists("embeddings_missing")) == 4
    assert results.index_size == 4
    assert set(sample_ids) == set(results.sample_ids)


def test_patches():
    dataset = _load_patches_dataset()

    results = dataset.load_brain_results("gt_sim")

    label_ids = dataset.values("ground_truth.detections.id", unwind=True)

    assert results.total_index_size == len(label_ids)
    assert set(label_ids) == set(results.label_ids)


def test_patches_subset():
    dataset = _load_patches_dataset()

    results = dataset.load_brain_results("gt_sim")

    label_ids = dataset.values("ground_truth.detections.id", unwind=True)

    assert results.total_index_size == len(label_ids)
    assert set(label_ids) == set(results.label_ids)

    view = dataset.filter_labels("ground_truth", F("label") == "person")
    results.use_view(view)

    label_ids = view.values("ground_truth.detections.id", unwind=True)

    assert results.index_size == len(label_ids)
    assert set(label_ids) == set(results.current_label_ids)


def test_patches_missing():
    dataset = _load_patches_dataset().limit(4).clone()
    dataset.add_samples(
        [
            fo.Sample(filepath="non-existent1.png"),
            fo.Sample(filepath="non-existent2.png"),
            fo.Sample(filepath="non-existent3.png"),
            fo.Sample(filepath="non-existent4.png"),
        ]
    )

    for sample in dataset[4:]:
        sample["ground_truth"] = fo.Detections(
            detections=[fo.Detection(bounding_box=[0.1, 0.1, 0.8, 0.8])]
        )
        sample.save()

    results = fob.compute_similarity(
        dataset, patches_field="ground_truth", batch_size=1
    )

    num_patches = dataset[:4].count("ground_truth.detections")
    label_ids = dataset[:4].values("ground_truth.detections.id", unwind=True)

    assert results.total_index_size == num_patches
    assert set(label_ids) == set(results.label_ids)

    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    results = fob.compute_similarity(
        dataset,
        model=model,
        patches_field="ground_truth",
        embeddings="embeddings_missing",
        batch_size=1,
    )

    view = dataset.filter_labels(
        "ground_truth", F("embeddings_missing") != None
    )

    assert view.count("ground_truth.detections") == num_patches
    assert results.total_index_size == num_patches
    assert set(label_ids) == set(results.label_ids)


def _load_images_dataset():
    name = "test-similarity-images"

    if fo.dataset_exists(name):
        return fo.load_dataset(name)

    return _make_images_dataset(name)


def _load_patches_dataset():
    name = "test-similarity-patches"

    if fo.dataset_exists(name):
        return fo.load_dataset(name)

    return _make_patches_dataset(name)


def _make_images_dataset(name):
    dataset = foz.load_zoo_dataset(
        "quickstart", max_samples=20, dataset_name=name
    )
    model = foz.load_zoo_model("inception-v3-imagenet-torch")

    # Embed images
    dataset.compute_embeddings(
        model, embeddings_field="embeddings", batch_size=8
    )

    # Image similarity
    fob.compute_similarity(
        dataset, embeddings="embeddings", brain_key="img_sim"
    )

    return dataset


def _make_patches_dataset(name):
    dataset = foz.load_zoo_dataset(
        "quickstart", max_samples=20, dataset_name=name
    )
    model = foz.load_zoo_model("inception-v3-imagenet-torch")

    # Embed ground truth patches
    dataset.compute_patch_embeddings(
        model,
        "ground_truth",
        embeddings_field="embeddings",
        batch_size=8,
        force_square=True,
    )

    # Patch similarity
    fob.compute_similarity(
        dataset,
        patches_field="ground_truth",
        embeddings="embeddings",
        brain_key="gt_sim",
    )

    return dataset


if __name__ == "__main__":
    fo.config.show_progress_bars = True
    unittest.main(verbosity=2)



"""
Similarity methods.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
import logging

from bson import ObjectId
import numpy as np
import sklearn.metrics as skm
import sklearn.neighbors as skn
import sklearn.preprocessing as skp

import eta.core.utils as etau

from fiftyone import ViewField as F
import fiftyone.core.brain as fob
import fiftyone.core.context as foc
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.patches as fop
import fiftyone.core.stages as fos
import fiftyone.core.validation as fov
import fiftyone.zoo as foz

from fiftyone.brain.similarity import SimilarityConfig, SimilarityResults
import fiftyone.brain.internal.core.utils as fbu


logger = logging.getLogger(__name__)

_AGGREGATIONS = {"mean": np.mean, "min": np.min, "max": np.max}

_DEFAULT_MODEL = "mobilenet-v2-imagenet-torch"
_DEFAULT_BATCH_SIZE = None

_MAX_PRECOMPUTE_DISTS = 15000  # ~1.7GB to store distance matrix in-memory
_COSINE_HACK_ATTR = "_cosine_hack"


def compute_similarity(
    samples,
    patches_field,
    embeddings,
    brain_key,
    metric,
    model,
    force_square,
    alpha,
    batch_size,
    num_workers,
    skip_failures,
):
    """See ``fiftyone/brain/__init__.py``."""

    fov.validate_collection(samples)

    if model is None and embeddings is None:
        model = _DEFAULT_MODEL
        if batch_size is None:
            batch_size = _DEFAULT_BATCH_SIZE

    if etau.is_str(model):
        _model = foz.load_zoo_model(model)
        try:
            supports_prompts = _model.can_embed_prompts
        except:
            supports_prompts = None
    else:
        _model = model
        supports_prompts = None

    if etau.is_str(embeddings):
        embeddings_field = fbu.parse_embeddings_field(
            samples,
            embeddings,
            patches_field=patches_field,
            allow_embedded=model is None,
        )
        embeddings = None
    else:
        embeddings_field = None

    config = SimilarityConfig(
        embeddings_field=embeddings_field,
        model=model,
        patches_field=patches_field,
        metric=metric,
        supports_prompts=supports_prompts,
    )
    brain_method = config.build()
    brain_method.ensure_requirements()

    if brain_key is not None:
        brain_method.register_run(samples, brain_key)

    embeddings, sample_ids, label_ids = fbu.get_embeddings(
        samples,
        model=_model,
        patches_field=patches_field,
        embeddings_field=embeddings_field,
        embeddings=embeddings,
        force_square=force_square,
        alpha=alpha,
        batch_size=batch_size,
        num_workers=num_workers,
        skip_failures=skip_failures,
    )

    results = SimilarityResults(
        samples,
        config,
        embeddings=embeddings,
        sample_ids=sample_ids,
        label_ids=label_ids,
    )

    brain_method.save_run_results(samples, brain_key, results)

    return results


class Similarity(fob.BrainMethod):
    def ensure_requirements(self):
        pass

    def get_fields(self, samples, brain_key):
        fields = []
        if self.config.patches_field is not None:
            fields.append(self.config.patches_field)

        return fields

    def cleanup(self, samples, brain_key):
        pass


class NeighborsHelper(object):
    def __init__(self, embeddings, metric):
        self.embeddings = embeddings
        self.metric = metric

        self._initialized = False
        self._full_dists = None
        self._curr_keep_inds = None
        self._curr_dists = None
        self._curr_neighbors = None

    def get_distances(self, keep_inds=None):
        self._init()

        if self._same_keep_inds(keep_inds):
            return self._curr_dists

        if keep_inds is not None:
            dists, _ = self._build(keep_inds=keep_inds, build_neighbors=False)
        else:
            dists = self._full_dists

        self._curr_keep_inds = keep_inds
        self._curr_dists = dists
        self._curr_neighbors = None

        return dists

    def get_neighbors(self, keep_inds=None):
        self._init()

        if self._curr_neighbors is not None and self._same_keep_inds(
            keep_inds
        ):
            return self._curr_neighbors, self._curr_dists

        dists, neighbors = self._build(keep_inds=keep_inds)

        self._curr_keep_inds = keep_inds
        self._curr_dists = dists
        self._curr_neighbors = neighbors

        return neighbors, dists

    def _same_keep_inds(self, keep_inds):
        if keep_inds is None and self._curr_keep_inds is None:
            return True

        if (
            isinstance(keep_inds, np.ndarray)
            and isinstance(self._curr_keep_inds, np.ndarray)
            and keep_inds.size == self._curr_keep_inds.size
            and (keep_inds == self._curr_keep_inds).all()
        ):
            return True

        return False

    def _init(self):
        if self._initialized:
            return

        # Pre-compute all pairwise distances if number of embeddings is small
        if len(self.embeddings) <= _MAX_PRECOMPUTE_DISTS:
            dists, _ = self._build_precomputed(
                self.embeddings, build_neighbors=False
            )
        else:
            dists = None

        self._initialized = True
        self._full_dists = dists
        self._curr_keep_inds = None
        self._curr_dists = dists
        self._curr_neighbors = None

    def _build(self, keep_inds=None, build_neighbors=True):
        # Use full distance matrix if available
        if self._full_dists is not None:
            if keep_inds is not None:
                dists = self._full_dists[keep_inds, :][:, keep_inds]
            else:
                dists = self._full_dists

            if build_neighbors:
                neighbors = skn.NearestNeighbors(metric="precomputed")
                neighbors.fit(dists)
            else:
                neighbors = None

            return dists, neighbors

        # Must build index
        embeddings = self.embeddings

        if keep_inds is not None:
            embeddings = embeddings[keep_inds]

        if len(embeddings) <= _MAX_PRECOMPUTE_DISTS:
            dists, neighbors = self._build_precomputed(
                embeddings, build_neighbors=build_neighbors
            )
        else:
            dists = None
            neighbors = self._build_graph(embeddings)

        return dists, neighbors

    def _build_precomputed(self, embeddings, build_neighbors=True):
        logger.info("Generating index...")

        # Center embeddings
        embeddings = np.asarray(embeddings)
        embeddings -= embeddings.mean(axis=0, keepdims=True)

        dists = skm.pairwise_distances(embeddings, metric=self.metric)

        if build_neighbors:
            neighbors = skn.NearestNeighbors(metric="precomputed")
            neighbors.fit(dists)
        else:
            neighbors = None

        logger.info("Index complete")

        return dists, neighbors

    def _build_graph(self, embeddings):
        logger.info(
            "Generating neighbors graph for %d embeddings; this may take "
            "awhile...",
            len(embeddings),
        )

        # Center embeddings
        embeddings = np.asarray(embeddings)
        embeddings -= embeddings.mean(axis=0, keepdims=True)

        metric = self.metric

        if metric == "cosine":
            # Nearest neighbors does not directly support cosine distance, so
            # we approximate via euclidean distance on unit-norm embeddings
            cosine_hack = True
            embeddings = skp.normalize(embeddings, axis=1)
            metric = "euclidean"
        else:
            cosine_hack = False

        neighbors = skn.NearestNeighbors(metric=metric)
        neighbors.fit(embeddings)

        setattr(neighbors, _COSINE_HACK_ATTR, cosine_hack)

        logger.info("Index complete")

        return neighbors


def plot_distances(results, bins, log, backend, **kwargs):
    _ensure_neighbors(results)

    keep_inds = results._curr_keep_inds
    neighbors, _ = results._neighbors_helper.get_neighbors(keep_inds=keep_inds)
    metric = results.config.metric
    thresh = results.thresh

    dists, _ = neighbors.kneighbors(n_neighbors=1)

    if backend == "matplotlib":
        return _plot_distances_mpl(dists, metric, thresh, bins, log, **kwargs)

    return _plot_distances_plotly(dists, metric, thresh, bins, log, **kwargs)


def sort_by_similarity(
    results, query, k, reverse, aggregation, dist_field, mongo
):
    _ensure_neighbors(results)

    samples = results.view
    sample_ids = results.current_sample_ids
    label_ids = results.current_label_ids
    keep_inds = results._curr_keep_inds
    patches_field = results.config.patches_field
    metric = results.config.metric

    selecting_samples = patches_field is None or isinstance(
        samples, fop.PatchesView
    )

    if aggregation not in _AGGREGATIONS:
        raise ValueError(
            "Unsupported aggregation method '%s'. Supported values are %s"
            % (aggregation, tuple(_AGGREGATIONS.keys()))
        )

    query = _parse_query(results, query)

    if isinstance(query, np.ndarray):
        # Query by vectors
        query_embeddings = query

        index_embeddings = results.embeddings
        if keep_inds is not None:
            index_embeddings = index_embeddings[keep_inds]

        dists = skm.pairwise_distances(
            index_embeddings, query_embeddings, metric=metric
        )
    else:
        # Query by IDs
        query_ids = query

        # Parse query IDs (always using full index)
        if patches_field is None:
            ids = results.sample_ids
        else:
            ids = results.label_ids

        bad_ids = []
        query_inds = []
        for query_id in query_ids:
            _inds = np.where(ids == query_id)[0]
            if _inds.size == 0:
                bad_ids.append(query_id)
            else:
                query_inds.append(_inds[0])

        if bad_ids:
            raise ValueError(
                "Query IDs %s were not included in this index" % bad_ids
            )

        # Perform query
        dists = results._neighbors_helper.get_distances()
        if dists is not None:
            # Use pre-computed distances
            if keep_inds is not None:
                dists = dists[keep_inds, :]

            dists = dists[:, query_inds]
        else:
            # Compute distances from embeddings
            index_embeddings = results.embeddings
            if keep_inds is not None:
                index_embeddings = index_embeddings[keep_inds]

            query_embeddings = results.embeddings[query_inds]
            dists = skm.pairwise_distances(
                index_embeddings, query_embeddings, metric=metric
            )

    agg_fcn = _AGGREGATIONS[aggregation]
    dists = agg_fcn(dists, axis=1)

    inds = np.argsort(dists)
    if reverse:
        inds = np.flip(inds)

    if k is not None:
        inds = inds[:k]

    if patches_field is not None:
        ids = label_ids
    else:
        ids = sample_ids

    #
    # Store query distances
    #

    if dist_field is not None:
        if selecting_samples:
            values = {ids[ind]: dists[ind] for ind in inds}
            samples.set_values(dist_field, values, key_field="id")
        else:
            label_type, path = samples._get_label_field_path(
                patches_field, dist_field
            )
            if issubclass(label_type, fol._LABEL_LIST_FIELDS):
                samples._set_list_values_by_id(
                    path,
                    sample_ids[inds],
                    label_ids[inds],
                    dists[inds],
                    path.rsplit(".", 1)[0],
                )
            else:
                values = {sample_ids[ind]: dists[ind] for ind in inds}
                samples.set_values(path, values, key_field="id")

    #
    # Construct sorted view
    #

    stages = []

    if selecting_samples:
        stage = fos.Select(ids[inds], ordered=True)
        stages.append(stage)
    else:
        # We're sorting by object similarity but this is not a patches view, so
        # arrange the samples in order of their first occuring label
        result_sample_ids = _unique_no_sort(sample_ids[inds])
        stage = fos.Select(result_sample_ids, ordered=True)
        stages.append(stage)

        if k is not None:
            _ids = [ObjectId(_id) for _id in label_ids[inds]]
            stage = fos.FilterLabels(patches_field, F("_id").is_in(_ids))
            stages.append(stage)

    if mongo:
        pipeline = []
        for stage in stages:
            stage.validate(samples)
            pipeline.extend(stage.to_mongo(samples))

        return pipeline

    view = samples
    for stage in stages:
        view = view.add_stage(stage)

    return view


def _parse_query(results, query):
    if isinstance(query, np.ndarray):
        # Query by vector(s)
        if query.size == 0:
            raise ValueError("At least one query vector must be provided")

        if query.ndim == 1:
            query = query[np.newaxis, :]

        return query

    if etau.is_str(query):
        query = [query]
    else:
        query = list(query)

    if not query:
        raise ValueError("At least one query must be provided")

    try:
        ObjectId(query[0])
        is_prompts = False
    except:
        is_prompts = True

    if is_prompts:
        if not results.config.supports_prompts:
            raise ValueError(
                "Invalid query ID '%s' (this model does not support prompts)"
                % query[0]
            )

        if results._model is None:
            results._model = foz.load_zoo_model(results.config.model)

        return results._model.embed_prompts(query)

    return query


def find_duplicates(results, thresh, fraction):
    _ensure_neighbors(results)

    keep_inds = results._curr_keep_inds
    embeddings = results.embeddings
    metric = results.config.metric
    patches_field = results.config.patches_field

    neighbors, dists = results._neighbors_helper.get_neighbors(
        keep_inds=keep_inds
    )

    if keep_inds is not None:
        embeddings = embeddings[keep_inds]

    if patches_field is not None:
        ids = results.current_label_ids
        logger.info("Computing duplicate patches...")
    else:
        ids = results.current_sample_ids
        logger.info("Computing duplicate samples...")

    num_embeddings = len(embeddings)

    #
    # Detect duplicates
    #

    if fraction is not None:
        num_keep = int(round(min(max(0, 1.0 - fraction), 1) * num_embeddings))
        keep, thresh = _remove_duplicates_count(
            neighbors,
            num_keep,
            num_embeddings,
            init_thresh=thresh,
        )
    else:
        keep = _remove_duplicates_thresh(neighbors, thresh, num_embeddings)

    unique_ids = [_id for idx, _id in enumerate(ids) if idx in keep]
    duplicate_ids = [_id for idx, _id in enumerate(ids) if idx not in keep]

    #
    # Locate nearest non-duplicate for each duplicate
    #

    if unique_ids and duplicate_ids:
        unique_inds = np.array(sorted(keep))
        dup_inds = np.array(
            [idx for idx in range(num_embeddings) if idx not in keep]
        )

        if dists is not None:
            # Use pre-computed distances
            _dists = dists[unique_inds, :][:, dup_inds]
            min_inds = np.argmin(_dists, axis=0)
            min_dists = _dists[min_inds, range(len(dup_inds))]
        else:
            neighbors = skn.NearestNeighbors(metric=metric)
            neighbors.fit(embeddings[unique_inds, :])
            min_dists, min_inds = neighbors.kneighbors(
                embeddings[dup_inds, :], n_neighbors=1
            )
            min_dists = min_dists.ravel()
            min_inds = min_inds.ravel()

        neighbors_map = defaultdict(list)
        for dup_id, min_ind, min_dist in zip(
            duplicate_ids, min_inds, min_dists
        ):
            nearest_id = ids[unique_inds[min_ind]]
            neighbors_map[nearest_id].append((dup_id, min_dist))

        neighbors_map = {
            k: sorted(v, key=lambda t: t[1]) for k, v in neighbors_map.items()
        }
    else:
        neighbors_map = {}

    results._thresh = thresh
    results._unique_ids = unique_ids
    results._duplicate_ids = duplicate_ids
    results._neighbors_map = neighbors_map

    logger.info("Duplicates computation complete")


def find_unique(results, count):
    _ensure_neighbors(results)

    keep_inds = results._curr_keep_inds
    neighbors, _ = results._neighbors_helper.get_neighbors(keep_inds=keep_inds)
    patches_field = results.config.patches_field
    num_embeddings = results.index_size

    if patches_field is not None:
        ids = results.current_label_ids
        logger.info("Computing unique patches...")
    else:
        ids = results.current_sample_ids
        logger.info("Computing unique samples...")

    # Find uniques
    keep, thresh = _remove_duplicates_count(neighbors, count, num_embeddings)

    unique_ids = [_id for idx, _id in enumerate(ids) if idx in keep]
    duplicate_ids = [_id for idx, _id in enumerate(ids) if idx not in keep]

    results._thresh = thresh
    results._unique_ids = unique_ids
    results._duplicate_ids = duplicate_ids
    results._neighbors_map = None

    logger.info("Uniqueness computation complete")


def duplicates_view(
    results, type_field, id_field, dist_field, sort_by, reverse
):
    samples = results.view
    patches_field = results.config.patches_field
    neighbors_map = results.neighbors_map

    if patches_field is not None and not isinstance(samples, fop.PatchesView):
        samples = samples.to_patches(patches_field)

    if sort_by == "distance":
        key = lambda kv: min(e[1] for e in kv[1])
    elif sort_by == "count":
        key = lambda kv: len(kv[1])
    else:
        raise ValueError(
            "Invalid sort_by='%s'; supported values are %s"
            % (sort_by, ("distance", "count"))
        )

    existing_ids = set(samples.values("id"))
    neighbors = [(k, v) for k, v in neighbors_map.items() if k in existing_ids]

    ids = []
    types = []
    nearest_ids = []
    dists = []
    for _id, duplicates in sorted(neighbors, key=key, reverse=reverse):
        ids.append(_id)
        types.append("nearest")
        nearest_ids.append(_id)
        dists.append(0)

        for dup_id, dist in duplicates:
            ids.append(dup_id)
            types.append("duplicate")
            nearest_ids.append(_id)
            dists.append(dist)

    dups_view = samples.select(ids, ordered=True)

    if type_field is not None:
        dups_view._dataset.add_sample_field(type_field, fof.StringField)
        dups_view.set_values(type_field, types)

    if id_field is not None:
        dups_view._dataset.add_sample_field(id_field, fof.StringField)
        dups_view.set_values(id_field, nearest_ids)

    if dist_field is not None:
        dups_view._dataset.add_sample_field(dist_field, fof.FloatField)
        dups_view.set_values(dist_field, dists)

    return dups_view


def unique_view(results):
    samples = results.view
    patches_field = results.config.patches_field
    unique_ids = results.unique_ids

    if patches_field is not None and not isinstance(samples, fop.PatchesView):
        samples = samples.to_patches(patches_field)

    return samples.select(unique_ids)


def values(results, path_or_expr):
    samples = results.view
    patches_field = results.config.patches_field
    if patches_field is not None:
        ids = results.current_label_ids
    else:
        ids = results.current_sample_ids

    return fbu.get_values(
        samples, path_or_expr, ids, patches_field=patches_field
    )


def visualize_duplicates(results, visualization, backend, **kwargs):
    visualization = _ensure_visualization(results, visualization)

    samples = results.view
    duplicate_ids = results.duplicate_ids
    neighbors_map = results.neighbors_map
    patches_field = results.config.patches_field

    dup_ids = set(duplicate_ids)
    nearest_ids = set(neighbors_map.keys())

    with visualization.use_view(samples, allow_missing=True):
        if patches_field is not None:
            ids = visualization.current_label_ids
        else:
            ids = visualization.current_sample_ids

        labels = []
        for _id in ids:
            if _id in dup_ids:
                label = "duplicate"
            elif _id in nearest_ids:
                label = "nearest"
            else:
                label = "unique"

            labels.append(label)

        if backend == "plotly":
            kwargs["edges"] = _build_edges(ids, neighbors_map)
            kwargs["edges_title"] = "neighbors"
            kwargs["labels_title"] = "type"

        return visualization.visualize(
            labels=labels,
            classes=["unique", "nearest", "duplicate"],
            backend=backend,
            **kwargs,
        )


def visualize_unique(results, visualization, backend, **kwargs):
    visualization = _ensure_visualization(results, visualization)

    samples = results.view
    unique_ids = results.unique_ids
    patches_field = results.config.patches_field

    unique_ids = set(unique_ids)

    with visualization.use_view(samples, allow_missing=True):
        if patches_field is not None:
            ids = visualization.current_label_ids
        else:
            ids = visualization.current_sample_ids

        labels = []
        for _id in ids:
            if _id in unique_ids:
                label = "unique"
            else:
                label = "other"

            labels.append(label)

        return visualization.visualize(
            labels=labels,
            classes=["other", "unique"],
            backend=backend,
            **kwargs,
        )


def _unique_no_sort(values):
    seen = set()
    return [v for v in values if v not in seen and not seen.add(v)]


def _ensure_neighbors(results):
    if results._neighbors_helper is not None:
        return

    embeddings = results.embeddings
    metric = results.config.metric
    results._neighbors_helper = NeighborsHelper(embeddings, metric)


def _ensure_visualization(results, visualization):
    if visualization is not None:
        return visualization

    import fiftyone.brain as fb

    samples = results._samples
    embeddings = results.embeddings
    patches_field = results.config.patches_field

    if embeddings.shape[1] in {2, 3}:
        config = fb.VisualizationConfig(
            patches_field=patches_field, num_dims=2
        )
        return fb.VisualizationResults(samples, config, embeddings)

    return fb.compute_visualization(
        samples,
        patches_field=patches_field,
        embeddings=embeddings,
        num_dims=2,
        seed=51,
        verbose=True,
    )


def _build_edges(ids, neighbors_map):
    inds_map = {_id: idx for idx, _id in enumerate(ids)}

    edges = []
    for nearest_id, duplicates in neighbors_map.items():
        nearest_ind = inds_map[nearest_id]
        for dup_id, _ in duplicates:
            dup_ind = inds_map[dup_id]
            edges.append((dup_ind, nearest_ind))

    return np.array(edges)


def _remove_duplicates_thresh(neighbors, thresh, num_embeddings):
    # When not using brute force, we approximate cosine distance by computing
    # Euclidean distance on unit-norm embeddings. ED = sqrt(2 * CD), so we need
    # to scale the threshold appropriately
    if getattr(neighbors, _COSINE_HACK_ATTR, False):
        thresh = np.sqrt(2.0 * thresh)

    inds = neighbors.radius_neighbors(radius=thresh, return_distance=False)

    keep = set(range(num_embeddings))
    for ind in range(num_embeddings):
        if ind in keep:
            keep -= {i for i in inds[ind] if i > ind}

    return keep


def _remove_duplicates_count(
    neighbors, num_keep, num_embeddings, init_thresh=None
):
    if init_thresh is not None:
        thresh = init_thresh
    else:
        thresh = 1

    if num_keep <= 0:
        logger.info("threshold: -, kept: %d, target: %d", num_keep, num_keep)
        return set(), None

    if num_keep >= num_embeddings:
        logger.info("threshold: -, kept: %d, target: %d", num_keep, num_keep)
        return set(range(num_embeddings)), None

    thresh_lims = [0, None]
    num_target = num_keep
    num_keep = -1

    while True:
        keep = _remove_duplicates_thresh(neighbors, thresh, num_embeddings)
        num_keep_last = num_keep
        num_keep = len(keep)

        logger.info(
            "threshold: %f, kept: %d, target: %d",
            thresh,
            num_keep,
            num_target,
        )

        if num_keep == num_target or (
            num_keep == num_keep_last
            and thresh_lims[1] is not None
            and thresh_lims[1] - thresh_lims[0] < 1e-6
        ):
            break

        if num_keep < num_target:
            # Need to decrease threshold
            thresh_lims[1] = thresh
            thresh = 0.5 * (thresh_lims[0] + thresh)
        else:
            # Need to increase threshold
            thresh_lims[0] = thresh
            if thresh_lims[1] is not None:
                thresh = 0.5 * (thresh + thresh_lims[1])
            else:
                thresh *= 2

    return keep, thresh


def _plot_distances_plotly(dists, metric, thresh, bins, log, **kwargs):
    import plotly.graph_objects as go
    import fiftyone.core.plots.plotly as fopl

    counts, edges = np.histogram(dists, bins=bins)
    left_edges = edges[:-1]
    widths = edges[1:] - edges[:-1]
    customdata = np.stack((edges[:-1], edges[1:]), axis=1)

    hover_lines = [
        "<b>count: %{y}</b>",
        "distance: [%{customdata[0]:.2f}, %{customdata[1]:.2f}]",
    ]
    hovertemplate = "<br>".join(hover_lines) + "<extra></extra>"

    bar = go.Bar(
        x=left_edges,
        y=counts,
        width=widths,
        customdata=customdata,
        offset=0,
        marker_color="#FF6D04",
        hovertemplate=hovertemplate,
        showlegend=False,
    )

    traces = [bar]

    if thresh is not None:
        line = go.Scatter(
            x=[thresh, thresh],
            y=[0, max(counts)],
            mode="lines",
            line=dict(color="#17191C", width=3),
            hovertemplate="<b>thresh: %{x}</b><extra></extra>",
            showlegend=False,
        )
        traces.append(line)

    figure = go.Figure(traces)

    figure.update_layout(
        xaxis_title="nearest neighbor distance (%s)" % metric,
        yaxis_title="count",
        hovermode="x",
        yaxis_rangemode="tozero",
    )

    if log:
        figure.update_layout(yaxis_type="log")

    figure.update_layout(**fopl._DEFAULT_LAYOUT)
    figure.update_layout(**kwargs)

    if foc.is_jupyter_context():
        figure = fopl.PlotlyNotebookPlot(figure)

    return figure


def _plot_distances_mpl(
    dists, metric, thresh, bins, log, ax=None, figsize=None, **kwargs
):
    import matplotlib.pyplot as plt

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    counts, edges = np.histogram(dists, bins=bins)
    left_edges = edges[:-1]
    widths = edges[1:] - edges[:-1]

    ax.bar(
        left_edges,
        counts,
        width=widths,
        align="edge",
        color="#FF6D04",
        **kwargs,
    )

    if thresh is not None:
        ax.vlines(thresh, 0, max(counts), color="#17191C", linewidth=3)

    if log:
        ax.set_yscale("log")

    ax.set_xlabel("nearest neighbor distance (%s)" % metric)
    ax.set_ylabel("count")

    if figsize is not None:
        fig.set_size_inches(*figsize)

    plt.tight_layout()

    return fig



"""
Trains a clean model using the brain.model simple_resnet class; it stores the
weights in a file for publishing the model.

Supports these config parameters:

    - config.take
    - config.epochs
    - config.batch_size
    - config.n_max
    - config.p_initial
    - config.n_rounds
    - config.cold_start
    - config.stats_path
    - config.model_path

A simple (minimal) set of these to run on a small machine is::

    run train_classifier.py -t 2000 -e 12 -b 64 --n_rounds 1 --p_initial 1.0 -m /tmp/foo.pth

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from functools import partial
import json
import os
import random
import sys
import time

from fiftyone.brain.internal.models.simple_resnet import *

from config import *
from datasets import *
from preprocess import *
from training import *
from utils import Timer


TEMP_TRAIN_DIR = "/tmp/le_test/train"
TEMP_VALID_DIR = "/tmp/le_test/valid"

localtime = lambda: time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


def main(config):

    ## Initial Data Input
    # Produces train_set and valid_set that are lists of tuples: (image, label)
    timer = Timer()
    whole_dataset = cifar10(root=DATA_DIR)
    print("Preprocessing training data")
    transforms = [
        partial(
            normalise,
            mean=np.array(cifar10_mean, dtype=np.float32),
            std=np.array(cifar10_std, dtype=np.float32),
        ),
        partial(transpose, source="NHWC", target="NCHW"),
    ]
    whole_train_set = list(
        zip(
            *preprocess(
                whole_dataset["train"], [partial(pad, border=4)] + transforms
            ).values()
        )
    )
    valid_set = list(
        zip(*preprocess(whole_dataset["valid"], transforms).values())
    )
    print(f"Finished loading and preprocessing in {timer():.2f} seconds")

    print(f"train set: {len(whole_train_set)} samples")
    print(f"valid set: {len(valid_set)} samples")

    if config.take:
        whole_train_set = whole_train_set[: config.take]
        valid_set = whole_train_set[: config.take]
        print(f"using a subset of the data")
        print(f"train set: {len(whole_train_set)} samples")
        print(f"valid set: {len(valid_set)} samples")

    # function of dataset
    N_labels = 10

    # set up the variables for training the model in each increment of the
    # dataset size
    lr_schedule = PiecewiseLinear([0, 5, config.epochs], [0, 0.4, 0])
    train_transforms = [Crop(32, 32), FlipLR(), Cutout(8, 8)]

    # compute the derived parameters for the trial based on the dataset and the
    # provided config.
    total_N = len(whole_train_set)

    # should be cleaner
    if config.n_max < 0:
        config.n_max = total_N

    start_N = round(config.p_initial * total_N)

    incr_N = (
        0
        if config.n_rounds == 1
        else round((config.n_max - start_N) / (config.n_rounds - 1))
    )

    print(f"Setting up the experiment: {total_N} training samples.")
    print(f"- starting with {start_N}")
    print(f"- incrementing by {incr_N} for each of {config.n_rounds-1} rounds")
    print(f"- total rounds: {config.n_rounds}")

    print(f"Starting the model training at {localtime()}")

    inuse_N = start_N

    model = simple_resnet().to(device).half()
    logs, state = Table(), {MODEL: model, LOSS: x_ent_loss}

    valid_batches = DataLoader(
        valid_set, config.batch_size, shuffle=False, drop_last=False
    )

    whole_train_set_use = whole_train_set[0:inuse_N]
    whole_train_set_avail = whole_train_set[inuse_N:]
    print(
        f"Split training set into two; "
        + f"using {len(whole_train_set_use)}, "
        + f"available {len(whole_train_set_avail)}"
    )

    stats = {}

    for iteration in range(config.n_rounds):
        print(f"beginning next round of training, using {inuse_N} samples")

        if config.cold_start:
            model = simple_resnet().to(device).half()
            logs, state = Table(), {MODEL: model, LOSS: x_ent_loss}

        train_batches = DataLoader(
            Transform(whole_train_set_use, train_transforms),
            config.batch_size,
            shuffle=True,
            set_random_choices=True,
            drop_last=True,
        )
        lr = lambda step: (
            lr_schedule(step / len(train_batches)) / config.batch_size
        )
        opts = [
            SGD(
                trainable_params(model).values(),
                {
                    "lr": lr,
                    "weight_decay": Const(5e-4 * config.batch_size),
                    "momentum": Const(0.9),
                },
            )
        ]
        state[OPTS] = opts

        for epoch in range(config.epochs):
            logs.append(
                union(
                    {"epoch": epoch + 1},
                    train_epoch(
                        state,
                        Timer(torch.cuda.synchronize),
                        train_batches,
                        valid_batches,
                    ),
                )
            )
        (
            logs.df()
            .query(f"epoch=={config.epochs}")[["train_acc", "valid_acc"]]
            .describe()
        )

        model.train(False)

        # record scores for this iteration
        iteration_stats = {}
        iteration_stats["in_use"] = inuse_N

        correct = 0
        total = 0
        class_correct = list(0.0 for i in range(10))
        class_total = list(0.0 for i in range(10))
        with torch.no_grad():
            for data in valid_batches.dataloader:
                images, labels = data
                y = model(images.cuda().half())
                _, predicted = torch.max(y, 1)
                total += labels.size(0)
                labels_gpu = labels.cuda().half()
                correct += (predicted == labels_gpu).sum().item()
                c = (predicted == labels_gpu).squeeze()
                for i in range(min(config.batch_size, len(labels))):
                    label = labels[i]
                    class_correct[label] += c[i].item()
                    class_total[label] += 1

        iteration_stats["validation_accuracy"] = correct / total

        model.train(True)

        # extend the corr_train_set_use with that from avail
        whole_train_set_use.extend(whole_train_set_avail[0:incr_N])
        whole_train_set_avail = whole_train_set_avail[incr_N:]
        inuse_N += incr_N
        assert inuse_N == len(whole_train_set_use)

        stats[inuse_N] = iteration_stats

    print(f"finished the full training; stats to follow")
    print(stats)

    if config.stats_path:
        with open(config.stats_path, "w") as fp:
            json.dump(stats, fp)

    if config.model_path:
        torch.save(model.state_dict(), config.model_path)


if __name__ == "__main__":

    config = commandline()

    print(f"running with config: {config}")

    main(config)



"""
Tests a model by loading the cifar10 dataset and then running predictions
against it.

@todo Update this to use the actual class functionality of the model rather
than the low-level model setup.

Run with at least the following command line:
    -m model_path.pth

Uses
    -t number

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from functools import partial
import os
import time

from imageio import imsave

import fiftyone as fo

from fiftyone.brain.internal.models.simple_resnet import *

from preprocess import *
from config import *
from datasets import *
from utils import Timer


TEMP_VALID_DIR = "/tmp/le_test/valid"

localtime = lambda: time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


def write_images(root, images, overwrite=False):
    paths = []
    for i, s in enumerate(images):
        path = os.path.join(root, f"{i:05d}.jpg")
        paths.append(path)

        if overwrite or not os.path.exists(path):
            img = s.copy()
            img = transpose(img, source="CHW", target="HWC")
            imsave(path, img)

    return paths


def main(config):

    ## Initial Data Input
    # Produces train_set and valid_set that are lists of tuples: (image, label)
    # (only the valid_set will be used in this script)
    timer = Timer()
    whole_dataset = cifar10(root=DATA_DIR)
    print("Preprocessing training data")
    transforms = [
        partial(
            normalise,
            mean=np.array(cifar10_mean, dtype=np.float32),
            std=np.array(cifar10_std, dtype=np.float32),
        ),
        partial(transpose, source="NHWC", target="NCHW"),
    ]
    valid_set = list(
        zip(*preprocess(whole_dataset["valid"], transforms).values())
    )
    print(f"Finished loading and preprocessing in {timer():.2f} seconds")

    print(f"valid set: {len(valid_set)} samples")

    if config.take:
        valid_set = valid_set[: config.take]
        print(f"using a subset of the data")
        print(f"valid set: {len(valid_set)} samples")

    # Write raw dataset to disk
    os.makedirs(TEMP_VALID_DIR, exist_ok=True)
    valid_image_paths = write_images(TEMP_VALID_DIR, list(zip(*valid_set))[0])

    # make the actual labels for the cifar-10 world
    labels = []
    for i, s in enumerate(cifar10_classes):
        labels.append(fo.Classification(label=s))

    #
    # Load data into FiftyOne
    #

    timer = Timer()
    dataset = fo.Dataset("le_cifar10")

    valid_samples = []
    for i, s in enumerate(valid_set):
        valid_samples.append(
            fo.Sample(
                valid_image_paths[i],
                tags=["valid"],
                ground_truth=fo.Classification(label=cifar10_classes[s[1]]),
            )
        )

    valid_ids = dataset.add_samples(valid_samples)

    print(f"Finished getting data into fiftyone in {timer():.2f} seconds")

    # load the model using the model path config
    assert config.model_path
    model = simple_resnet().to(device).half()
    model.load_state_dict(torch.load(config.model_path))

    print("Model loaded.")

    model.train(False)

    # I need to get my datasets into a format where I'll have the ids available
    # as well during the data loading

    valid_imgs, valid_labels = zip(*valid_set)
    fo_valid_set = list(zip(valid_imgs, valid_labels, valid_ids))

    valid_batches = DataLoader(
        fo_valid_set, config.batch_size, shuffle=False, drop_last=False
    )

    print("running predictions on the data")

    correct = 0
    total = 0
    class_correct = list(0.0 for i in range(10))
    class_total = list(0.0 for i in range(10))
    with torch.no_grad():
        for images, labels, ids in valid_batches.dataloader:
            y = model(images.cuda().half())
            _, predicted = torch.max(y, 1)
            total += labels.size(0)
            labels_gpu = labels.cuda().half()
            correct += (predicted == labels_gpu).sum().item()
            c = (predicted == labels_gpu).squeeze()
            for i in range(min(config.batch_size, len(labels))):
                label = labels[i]
                class_correct[label] += c[i].item()
                class_total[label] += 1

            for prediction, theid, thelogit in zip(predicted, ids, y):
                label = fo.Classification(
                    label=cifar10_classes[prediction], logits=thelogit
                )
                dataset[theid]["prediction"] = label
                dataset[theid]["max-logit"] = float(
                    np.max(thelogit.cpu().numpy())
                )

    print(
        "Accuracy of the network on the 10K test images: %.2f%%"
        % (100 * correct / total)
    )

    for i in range(10):
        print(
            "Accuracy of %9s : %.2f%%"
            % (cifar10_classes[i], 100 * class_correct[i] / class_total[i])
        )

    print("done")

    return dataset


if __name__ == "__main__":

    config = commandline()

    print(f"running with config: {config}")

    dataset = main(config)
